#!/usr/bin/python
#
# Copyright 2023 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

# NOTE: THIS FILE IS OBSOLETE AND WILL BE REMOVED IN 2025H2

import kaggle_api_extended as api
import json
import os
import re  # noqa: F401
import tempfile


class KaggleApi(object):

  def __init__(self, api_client=None):
    if api_client is None:
      api_client = api.KaggleApi()
    self.api_client = api_client

  def competition_download_leaderboard(self, id):  # noqa: E501
    """Download competition leaderboard  # noqa: E501.

    :param str id: Competition name (required)
    :return: Result
    """
    return self.api_client.competition_leaderboard_download(id, None)

  def competition_view_leaderboard(self, id):  # noqa: E501
    """View competition leaderboard  # noqa: E501.

    :param str id: Competition name (required)
    :return: Result
    """
    return self.api_client.competition_leaderboard_view(id)

  def competitions_data_download_file(self, id, file_name):  # noqa: E501
    """Download competition data file  # noqa: E501.

    :param str id: Competition name (required)
    :param str file_name: Competition name (required)
    :return: Result
    """
    return self.api_client.competition_download_file(id, file_name)

  def competitions_data_download_files(self, id):  # noqa: E501
    """Download all competition data files  # noqa: E501.

    :param str id: Competition name (required)
    :return: Result
    """
    return self.api_client.competition_download_files(id)

  def competitions_data_list_files(self, id):  # noqa: E501
    """List competition data files  # noqa: E501.

    :param str id: Competition name (required)
    :return: Result
    """
    return self.api_client.competition_list_files(id)

  def competitions_list(self,
                        group=None,
                        category=None,
                        sort_by=None,
                        page=1,
                        search=None):  # noqa: E501
    """List competitions  # noqa: E501.

    :param str group: Filter competitions by a particular group
    :param str category: Filter competitions by a particular category
    :param str sort_by: Sort the results
    :param int page: Page number
    :param str search: Search terms
    :return: ApiListCompetitionsResponse
    """
    return self.api_client.competitions_list(group, category, sort_by, page,
                                             search)

  def competitions_submissions_list(self, id, page=0):  # noqa: E501
    """List competition submissions  # noqa: E501.

    :param str id: Competition name (required)
    :param int page: Page number
    :return: Result
    """
    return self.api_client.competition_submissions(id, page)

  def competitions_submissions_submit(self, blob_file_tokens,
                                      submission_description, id):  # noqa: E501
    """Submit to competition  # noqa: E501.

    :param str blob_file_tokens: Token identifying location of uploaded
        submission file (required)
    :param str submission_description: Description of competition
        submission (required)
    :param str id: Competition name (required)
    :return: Result
    """
    print("competitions_submissions_submit() not implemented")
    print("use kaggle_api_extended.KaggleApi.competition_submit() instead")
    raise NotImplementedError()

  def competitions_submissions_upload(self, file, guid, content_length,
                                      last_modified_date_utc):  # noqa: E501
    """Upload competition submission file  # noqa: E501.

    :param file file: Competition submission file (required)
    :param str guid: Location where submission should be uploaded
        (required)
    :param int content_length: Content length of file in bytes
        (required)
    :param int last_modified_date_utc: Last modified date of file in
        seconds since epoch in UTC (required)
    :return: Result
    """
    print("competitions_submissions_upload() not implemented")
    print(
        "use kaggle_api_extended.KaggleApi.competitions_submissions_submit() instead"
    )
    raise NotImplementedError()

  def competitions_submissions_url(self, id, content_length,
                                   last_modified_date_utc):  # noqa: E501
    """Generate competition submission URL  # noqa: E501.

    :param str id: Competition name, as it appears in the competition's
        URL (required)
    :param int content_length: Content length of file in bytes
        (required)
    :param int last_modified_date_utc: Last modified date of file in
        seconds since epoch in UTC (required)
    :param str file_name: Competition submission file name
    :return: Result
    """
    print("competitions_submissions_url() not implemented")
    print("use kaggle_api_extended.KaggleApi.competition_submit() instead")
    raise NotImplementedError()

  def create_inbox_file(self, create_inbox_file_request):  # noqa: E501
    """Creates (aka \&quot;drops\&quot;) a new file into the inbox.  # noqa:
    E501.

    :param CreateInboxFileRequest create_inbox_file_request: (required)
    :return: Result
    """
    print("create_inbox_file() not implemented")
    print("use kaggle_api_extended.KaggleApi.files_upload_cli() instead")
    raise NotImplementedError()

  def datasets_create_new(self, request):  # noqa: E501
    """Create a new dataset  # noqa: E501.

    :param DatasetNewRequest request: Information for creating a new
        dataset (required)
    :return: Result
    """
    print("datasets_create_new() not implemented")
    print("use kaggle_api_extended.KaggleApi.dataset_create_new() instead")
    raise NotImplementedError()

  def datasets_create_version_by_id(self, id, dataset_new_version_request,
                                    **kwargs):  # noqa: E501
    """Create a new dataset version by id  # noqa: E501.

    :param int id: Dataset ID (required)
    :param DatasetNewVersionRequest dataset_new_version_request:
        Information for creating a new dataset version (required)
    :return: throws NotImplementedError
    """
    print("datasets_create_version_by_id() not implemented")
    print("use kaggle_api_extended.KaggleApi.dataset_create_version() instead")
    raise NotImplementedError()

  def datasets_download(self,
                        owner_slug,
                        dataset_slug,
                        dataset_version_number=None):  # noqa: E501
    """Download dataset file  # noqa: E501.

    :param str owner_slug: Dataset owner (required)
    :param str dataset_slug: Dataset name (required)
    :param str dataset_version_number: Dataset version number
    :return: Result
    """
    dataset = f'{owner_slug}/{dataset_slug}'
    if dataset_version_number is not None:
      dataset += f'/{dataset_version_number}'
    return self.api_client.dataset_download_files(dataset)

  def datasets_download_file(self,
                             owner_slug,
                             dataset_slug,
                             file_name,
                             dataset_version_number=None):  # noqa: E501
    """Download dataset file  # noqa: E501.

    :param str owner_slug: Dataset owner (required)
    :param str dataset_slug: Dataset name (required)
    :param str file_name: File name (required)
    :param str dataset_version_number: Dataset version number
    :return: True if the file was downloaded successfully, False otherwise
    """
    dataset = f'{owner_slug}/{dataset_slug}'
    if dataset_version_number is not None:
      dataset += f'@{dataset_version_number}'
    return self.api_client.dataset_download_file(dataset, file_name)

  def datasets_list(self,
                    sort_by=None,
                    size=None,
                    file_type=None,
                    license_name=None,
                    tag_ids=None,
                    search=None,
                    user=None,
                    mine=False,
                    page=1,
                    max_size=None,
                    min_size=None):  # noqa: E501
    """List datasets  # noqa: E501.

    :param str group: Display datasets by a particular group
    :param str sort_by: Sort the results
    :param str size: (DEPRECATED). Please use --max-size and --min-size
        to filter dataset sizes.
    :param str filetype: Display datasets of a specific file type
    :param str license: Display datasets with a specific license
    :param str tagids: A comma separated list of tags to filter by
    :param str search: Search terms
    :param str user: Display datasets by a specific user or organization
    :param int page: Page number
    :param int max_size: Max Dataset Size (bytes)
    :param int min_size: Max Dataset Size (bytes)
    :return: List of ApiDatasetFile.
    """
    return self.api_client.dataset_list(
        sort_by=sort_by,
        size=size,
        file_type=file_type,
        license_name=license_name,
        tag_ids=tag_ids,
        search=search,
        user=user,
        mine=mine,
        page=page,
        max_size=max_size,
        min_size=min_size)

  def datasets_status(self, owner_slug, dataset_slug, **kwargs):  # noqa: E501
    """Get dataset creation status  # noqa: E501.

    :param str owner_slug: Dataset owner (required)
    :param str dataset_slug: Dataset name (required)
    :return: Result
    """
    return self.api_client.dataset_status(f'{owner_slug}/{dataset_slug})')

  def delete_model(self, owner_slug, model_slug):  # noqa: E501
    """Delete a model  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model name (required)
    :return: ApiDeleteModelResponse
    """
    return self.api_client.model_delete(f'{owner_slug}/{model_slug})')

  def delete_model_instance(self, owner_slug, model_slug, framework,
                            instance_slug):  # noqa: E501
    """Delete a model instance  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model name (required)
    :param str framework: Model instance framework (required)
    :param str instance_slug: Model instance slug (required)
    :return: Result
    """
    return self.api_client.model_instance_delete(
        f'{owner_slug}/{model_slug}/{framework}/{instance_slug}', yes=True)

  def delete_model_instance_version(self, owner_slug, model_slug, framework,
                                    instance_slug,
                                    version_number):  # noqa: E501
    """Delete a model instance version  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model name (required)
    :param str framework: Model instance framework (required)
    :param str instance_slug: Model instance slug (required)
    :param str version_number: Model instance version number (required)
    :return: Result
    """
    return self.api_client.model_instance_version_delete(
        f'{owner_slug}/{model_slug}/{framework}/{instance_slug}/{version_number}',
        yes=True)

  def get_model(self, owner_slug, model_slug):  # noqa: E501
    """Get a model  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model name (required)
    :return: Result
    """
    return self.api_client.model_get(f'{owner_slug}/{model_slug}')

  def get_model_instance(self, owner_slug, model_slug, framework,
                         instance_slug):  # noqa: E501
    """Get a model instance  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model name (required)
    :param str framework: Model instance framework (required)
    :param str instance_slug: Model instance slug (required)
    :return: Result
    """
    return self.api_client.model_instance_get(
        f'{owner_slug}/{model_slug}/{framework}/{instance_slug}')

  def kernel_output(self, user_name, kernel_slug):  # noqa: E501
    """Download the latest output from a kernel  # noqa: E501.

    :param str user_name: Kernel owner (required)
    :param str kernel_slug: Kernel name (required)
    :return: Result
    """
    return self.api_client.kernels_output(
        f'{user_name}/{kernel_slug}', path=None, force=True)

  def kernel_pull(self, user_name, kernel_slug):  # noqa: E501
    """Pull the latest code from a kernel  # noqa: E501.

    :param str user_name: Kernel owner (required)
    :param str kernel_slug: Kernel name (required)
    :return: Result
    """
    return self.api_client.kernels_pull(f'{user_name}/{kernel_slug}', path=None)

  def kernel_push(self, kernel_push_request):  # noqa: E501
    """Push a new kernel version.  Can be used to create a new kernel and
    update an existing one.  # noqa: E501.

    :param KernelPushRequest kernel_push_request: Information for
        pushing a new kernel version (required)
    :return: Result
    """
    with tempfile.TemporaryDirectory() as tmpdir:
      meta_file = os.path.join(tmpdir, 'kernel-metadata.json')
      (fd, code_file) = tempfile.mkstemp('code', 'py', tmpdir, text=True)
      fd.write(json.dumps(kernel_push_request.code))
      os.close(fd)
      with open(meta_file, 'w') as f:
        params = kernel_push_request.to_dict()
        params['code_file'] = code_file
        params['competition_sources'] = params.get('competition_data_sources')
        params['dataset_sources'] = params.get('dataset_data_sources')
        params['kernel_sources'] = params.get('kernel_data_sources')
        params['model_sources'] = params.get('model_data_sources')
        params['title'] = params.get('new_title')
        entries_to_remove = ('competition_data_sources', 'dataset_data_sources',
                             'kernel_data_sources', 'model_data_sources',
                             'new_title')
        for k in entries_to_remove:
          params.pop(k, None)
        f.write(json.dumps(params))
      return self.api_client.kernels_push(tmpdir)

  def kernel_status(self, user_name, kernel_slug):  # noqa: E501
    """Get the status of the latest kernel version  # noqa: E501.

    :param str user_name: Kernel owner (required)
    :param str kernel_slug: Kernel name (required)
    :return: Result
    """
    return self.api_client.kernels_status(f'{user_name}/{kernel_slug}')

  def kernels_list(self,
                   page=1,
                   page_size=20,
                   dataset=None,
                   competition=None,
                   parent_kernel=None,
                   search=None,
                   group='everyone',
                   user=None,
                   language=None,
                   kernel_type=None,
                   output_type=None,
                   sort_by=None):  # noqa: E501
    """List kernels  # noqa: E501.

    :param int page: Page number
    :param int page_size: Page size
    :param str search: Search terms
    :param str group: Display only your kernels
    :param str user: Display kernels by a particular group
    :param str language: Display kernels in a specific language
    :param str kernel_type: Display kernels of a specific type
    :param str output_type: Display kernels with a specific output type
    :param str sort_by: Sort the results. 'relevance' only works if
        there is a search query
    :param str dataset: Display kernels using the specified dataset
    :param str competition: Display kernels using the specified
        competition
    :param str parent_kernel: Display kernels that have forked the
        specified kernel
    :return: Result
    """
    return self.api_client.kernels_list(
        page=page,
        page_size=page_size,
        dataset=dataset,
        competition=competition,
        parent_kernel=parent_kernel,
        search=search,
        mine=group != 'everyone',
        user=user,
        language=language,
        kernel_type=kernel_type,
        output_type=output_type,
        sort_by=sort_by)

  def metadata_get(self, owner_slug, dataset_slug):  # noqa: E501
    """Get the metadata for a dataset  # noqa: E501.

    :param str owner_slug: Dataset owner (required)
    :param str dataset_slug: Dataset name (required)
    :return: Result
    """
    return self.api_client.dataset_metadata(f'{owner_slug}/{dataset_slug}',
                                            None)

  def metadata_post(self, owner_slug, dataset_slug, settings,
                    request):  # noqa: E501
    """Update the metadata for a dataset  # noqa: E501.

    :param str owner_slug: Dataset owner (required)
    :param str dataset_slug: Dataset name (required)
    :param DatasetUpdateSettingsRequest settings: Dataset metadata to
        update (required)
    :return: Result
    """
    with tempfile.TemporaryDirectory() as tmpdir:
      meta_file = os.path.join(tmpdir, 'dataset-metadata.json')
      with open(meta_file, 'w') as f:
        params = request.to_dict()
        params['isPrivate'] = params.get('is_private')
        params.pop('is_private', None)
        f.write(json.dumps(params))
      return self.api_client.dataset_metadata_update(
          f'{owner_slug}/{dataset_slug}', meta_file)

  def model_instance_versions_download(self, owner_slug, model_slug, framework,
                                       instance_slug,
                                       version_number):  # noqa: E501
    """Download model instance version files  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model name (required)
    :param str framework: Model instance framework (required)
    :param str instance_slug: Model instance slug (required)
    :param str version_number: Model instance version number (required)
    :return: Result
    """
    v = f'{owner_slug}/{model_slug}/{framework}/{instance_slug}/{version_number}'
    return self.api_client.model_instance_version_download(v)

  def models_create_instance(self, owner_slug, model_slug,
                             model_new_instance_request):  # noqa: E501
    """Create a new model instance  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model slug (required)
    :param ModelNewInstanceRequest model_new_instance_request:
        Information for creating a new model instance (required)
    :return: Result
    """
    print("models_create_instance() not implemented")
    print("use kaggle_api_extended.KaggleApi.model_instance_create() instead")
    raise NotImplementedError()

  def models_create_instance_version(
      self, owner_slug, model_slug, framework, instance_slug,
      model_instance_new_version_request):  # noqa: E501
    """Create a new model instance version  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model slug (required)
    :param str framework: Model instance framework (required)
    :param str instance_slug: Model instance slug (required)
    :param ModelInstanceNewVersionRequest
        model_instance_new_version_request: Information for creating a
        new model instance version (required)
    :return: Result
    """
    print("models_create_instance_version() not implemented")
    print(
        "use kaggle_api_extended.KaggleApi.model_instance_version_create() instead"
    )
    raise NotImplementedError()

  def models_create_new(self, model_new_request):  # noqa: E501
    """Create a new model  # noqa: E501.

    :param ModelNewRequest model_new_request: Information for creating a
        new model (required)
    :return: Result
    """
    with tempfile.TemporaryDirectory() as tmpdir:
      meta_file = os.path.join(tmpdir, 'model-metadata.json')
      with open(meta_file, 'w') as f:
        params = model_new_request.to_dict()
        params['ownerSlug'] = params.get('owner_slug')
        params['isPrivate'] = params.get('is_private')
        params['publishTime'] = params.get('publish_time')
        params['provenanceSources'] = params.get('provenance_sources')
        entries_to_remove = ('owner_slug', 'is_private', 'publish_time',
                             'provenance_sources')
        for k in entries_to_remove:
          params.pop(k, None)
        f.write(json.dumps(params))
      return self.api_client.model_create_new(tmpdir)

  def models_list(self,
                  sort_by=None,
                  search=None,
                  owner=None,
                  page_size=20,
                  page_token=None):
    """List models
    
    :param str search: Search terms
    :param str sort_by: Sort the results
    :param str owner: Display models by a specific user or organization
    :param int page_size: Page size
    :param str page_token: Page token for pagination
    :return: list of models
    """
    return self.api_client.model_list(
        sort_by=sort_by,
        search=search,
        owner=owner,
        page_size=page_size,
        page_token=page_token)

  def update_model(
      self,
      owner_slug,
      model_slug,
      model_update_request,
  ):  # noqa: E501
    """Update a model  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model name (required)
    :param ModelUpdateRequest model_update_request: Information for
        updating a model (required)
    :return: Result
    """
    with tempfile.TemporaryDirectory() as tmpdir:
      meta_file = os.path.join(tmpdir, 'model-metadata.json')
      with open(meta_file, 'w') as f:
        params = model_update_request.to_dict()
        params['ownerSlug'] = params.get('owner_slug')
        params['isPrivate'] = params.get('is_private')
        params['publishTime'] = params.get('publish_time')
        params['provenanceSources'] = params.get('provenance_sources')
        params['updateMask'] = params.get('update_mask')
        entries_to_remove = ('owner_slug', 'is_private', 'publish_time',
                             'provenance_sources', 'update_mask')
        for k in entries_to_remove:
          params.pop(k, None)
        f.write(json.dumps(params))
      return self.api_client.model_update(tmpdir)

  def update_model_instance(self, owner_slug, model_slug, framework,
                            instance_slug,
                            model_instance_update_request):  # noqa: E501
    """Update a model  # noqa: E501.

    :param str owner_slug: Model owner (required)
    :param str model_slug: Model name (required)
    :param str framework: Model instance framework (required)
    :param str instance_slug: Model instance slug (required)
    :param ModelInstanceUpdateRequest model_instance_update_request:
        Information for updating a model instance (required)
    :return: Result
    """
    with tempfile.TemporaryDirectory() as tmpdir:
      meta_file = os.path.join(tmpdir, 'model-instance-metadata.json')
      with open(meta_file, 'w') as f:
        params = model_instance_update_request.to_dict()
        params['licenseName'] = params.get('license_name')
        params['fineTunable'] = params.get('fine_tunable')
        params['trainingData'] = params.get('training_data')
        params['modelInstanceType'] = params.get('model_instance_type')
        params['baseModelInstance'] = params.get('base_model_instance')
        params['externalBaseModelUrl'] = params.get('external_base_model_url')
        params['updateMask'] = params.get('update_mask')
        entries_to_remove = ('license_name', 'fine_tunable', 'training_data',
                             'model_instance_type', 'base_model_instance',
                             'external_base_model_url', 'update_mask')
        for k in entries_to_remove:
          params.pop(k, None)
        f.write(json.dumps(params))
      return self.api_client.model_instance_update(tmpdir)

  def upload_file(self, start_blob_upload_request):  # noqa: E501
    """Start uploading a file  # noqa: E501 :param StartBlobUploadRequest
    start_blob_upload_request: (required) :return: StartBlobUploadResponse."""
    print("upload_file() not implemented")
    print("it should not be needed after converting your code")
    print("to use kaggle_api_extended.KaggleApi.<high-level-fn>() instead")
    raise NotImplementedError()



# coding=utf-8
from setuptools import setup, find_packages

# Note: pyproject.toml seems to be chosen by pip install over setup.py, so this
# file should not be used anymore. However, cloudbuild.yaml still uses setup.py
# to drive the build that is published to pypi.org. That needs to be changed
# before this file can be removed. And, due to that, pyproject.toml needs to be
# deleted before a release can be made.
# https://packaging.python.org/en/latest/guides/modernize-setup-py-project/
setup(
    name='kaggle',
    version='1.7.4.2',
    description='Kaggle API',
    long_description=(
        'Official API for https://www.kaggle.com, accessible using a command line '
        'tool implemented in Python. Beta release - Kaggle reserves the right to '
        'modify the API functionality currently offered.'),
    author='Kaggle',
    author_email='support@kaggle.com',
    url='https://github.com/Kaggle/kaggle-api',
    project_urls={
        'Documentation': 'https://www.kaggle.com/docs/api',
        'GitHub': 'https://github.com/Kaggle/kaggle-api',
        'Tracker': 'https://github.com/Kaggle/kaggle-api/issues',
    },
    keywords=['Kaggle', 'API'],
    entry_points={'console_scripts': ['kaggle = kaggle.cli:main']},
    install_requires=[
        'six >= 1.10', 'certifi >= 2023.7.22', 'python-dateutil', 'requests',
        'tqdm', 'python-slugify', 'urllib3', 'bleach', 'protobuf',
        'hatchling >= 1.27.0'
    ],
    packages=find_packages(
        where='src',
        include=['kaggle*'],
    ),
    package_dir={"": "src"})



from kagglesdk import kaggle_env
from kagglesdk import KaggleClient, KaggleEnv

# python -m unittest tests.test_authenticate

import os
import unittest


class TestClient(unittest.TestCase):

  def setUp(self):
    print('setup             class:%s' % self)

  def tearDown(self):
    print('teardown          class:TestStuff')

  # Environment

  def test_kaggle_environment(self):
    os.environ['KAGGLE_API_ENVIRONMENT'] = 'PROD'

    env = kaggle_env.get_env()
    self.assertEqual(env, KaggleEnv.PROD)

    endpoint = kaggle_env.get_endpoint(env)
    self.assertEqual(endpoint, 'https://www.kaggle.com')

  # Client

  def test_kaggle_client(self):
    client = KaggleClient(
        env=KaggleEnv.PROD, verbose=False, username='dinosaur', password='xxxxxxxxxxxx'
    )

    self.assertEqual(client.username, 'dinosaur')
    self.assertEqual(client.password, 'xxxxxxxxxxxx')
    self.assertEqual(client.http_client()._endpoint, 'https://www.kaggle.com')
    self.assertEqual(client.http_client()._verbose, False)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/python
#
# Copyright 2019 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
from __future__ import print_function

import argparse
import json

import six

from kaggle import KaggleApi
from kaggle import api
# from rest import ApiException
ApiException = IOError


def main():
  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawTextHelpFormatter)

  parser.add_argument(
      '-v',
      '--version',
      action='version',
      help='Print the Kaggle API version',
      version='Kaggle API ' + KaggleApi.__version__)
  parser.add_argument(
      '-W',
      '--no-warn',
      dest='disable_version_warning',
      action='store_true',
      help='Disable out-of-date API version warning')

  subparsers = parser.add_subparsers(
      title='commands', help=Help.kaggle, dest='command')
  subparsers.required = True
  subparsers.choices = Help.kaggle_choices
  parse_competitions(subparsers)
  parse_datasets(subparsers)
  parse_kernels(subparsers)
  parse_models(subparsers)
  parse_files(subparsers)
  parse_config(subparsers)
  args = parser.parse_args()
  command_args = {}
  command_args.update(vars(args))
  del command_args['func']
  del command_args['command']
  if command_args['disable_version_warning']:
    KaggleApi.already_printed_version_warning = True
  del command_args['disable_version_warning']
  error = False
  try:
    out = args.func(**command_args)
  except ApiException as e:
    print(e)
    out = None
    error = True
  except ValueError as e:
    print(e)
    out = None
    error = True
  except KeyboardInterrupt:
    print('User cancelled operation')
    out = None
  if out is not None:
    print(out, end='')

  # This is so that scripts that pick up on error codes can tell when there was a failure
  if error:
    exit(1)


def __parse_body(body):
  try:
    return json.loads(body)
  except Exception as e:
    return {}


def parse_competitions(subparsers):
  if six.PY2:
    parser_competitions = subparsers.add_parser(
        'competitions',
        formatter_class=argparse.RawTextHelpFormatter,
        help=Help.group_competitions)
  else:
    parser_competitions = subparsers.add_parser(
        'competitions',
        formatter_class=argparse.RawTextHelpFormatter,
        help=Help.group_competitions,
        aliases=['c'])
  subparsers_competitions = parser_competitions.add_subparsers(
      title='commands', dest='command')
  subparsers_competitions.required = True
  subparsers_competitions.choices = Help.competitions_choices

  # Competitions list
  parser_competitions_list = subparsers_competitions.add_parser(
      'list',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_competitions_list)
  parser_competitions_list_optional = parser_competitions_list._action_groups.pop(
  )
  parser_competitions_list_optional.add_argument(
      '--group',
      dest='group',
      required=False,
      help=Help.param_competition_group)
  parser_competitions_list_optional.add_argument(
      '--category',
      dest='category',
      required=False,
      help=Help.param_competition_category)
  parser_competitions_list_optional.add_argument(
      '--sort-by',
      dest='sort_by',
      required=False,
      help=Help.param_competition_sort_by)
  parser_competitions_list_optional.add_argument(
      '-p',
      '--page',
      dest='page',
      default=1,
      type=int,
      required=False,
      help=Help.param_page)
  parser_competitions_list_optional.add_argument(
      '-s', '--search', dest='search', required=False, help=Help.param_search)
  parser_competitions_list_optional.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_competitions_list._action_groups.append(
      parser_competitions_list_optional)
  parser_competitions_list.set_defaults(func=api.competitions_list_cli)

  # Competitions list files
  parser_competitions_files = subparsers_competitions.add_parser(
      'files',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_competitions_files)
  parser_competitions_files_optional = parser_competitions_files._action_groups.pop(
  )
  parser_competitions_files_optional.add_argument(
      'competition', nargs='?', default=None, help=Help.param_competition)
  parser_competitions_files_optional.add_argument(
      '-c',
      '--competition',
      dest='competition_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_competitions_files_optional.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_competitions_files_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_competitions_files_optional.add_argument(
      '--page-token',
      dest='page_token',
      required=False,
      help=Help.param_page_token)
  parser_competitions_files_optional.add_argument(
      '--page-size',
      dest='page_size',
      required=False,
      default=20,
      type=int,
      help=Help.param_page_size)
  parser_competitions_files._action_groups.append(
      parser_competitions_files_optional)
  parser_competitions_files.set_defaults(func=api.competition_list_files_cli)

  # Competitions download
  parser_competitions_download = subparsers_competitions.add_parser(
      'download',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_competitions_download)
  parser_competitions_download_optional = parser_competitions_download._action_groups.pop(
  )
  parser_competitions_download_optional.add_argument(
      'competition', nargs='?', default=None, help=Help.param_competition)
  parser_competitions_download_optional.add_argument(
      '-c',
      '--competition',
      dest='competition_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_competitions_download_optional.add_argument(
      '-f',
      '--file',
      dest='file_name',
      required=False,
      help=Help.param_competition_file)
  parser_competitions_download_optional.add_argument(
      '-p', '--path', dest='path', required=False, help=Help.param_downfolder)
  parser_competitions_download_optional.add_argument(
      '-w',
      '--wp',
      dest='path',
      action='store_const',
      const='.',
      required=False,
      help=Help.param_wp)
  parser_competitions_download_optional.add_argument(
      '-o', '--force', dest='force', action='store_true', help=Help.param_force)
  parser_competitions_download_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_competitions_download._action_groups.append(
      parser_competitions_download_optional)
  parser_competitions_download.set_defaults(func=api.competition_download_cli)

  # Competitions submit
  parser_competitions_submit = subparsers_competitions.add_parser(
      'submit',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_competitions_submit)
  parser_competitions_submit_optional = parser_competitions_submit._action_groups.pop(
  )
  parser_competitions_submit_optional.add_argument(
      'competition', nargs='?', default=None, help=Help.param_competition)
  parser_competitions_submit_optional.add_argument(
      '-c',
      '--competition',
      dest='competition_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_competitions_submit_optional.add_argument(
      '-f', '--file', dest='file_name', help=Help.param_upfile)
  parser_competitions_submit_optional.add_argument(
      '-k', '--kernel', dest='kernel', help=Help.param_code_kernel)
  parser_competitions_submit_optional.add_argument(
      '-m',
      '--message',
      dest='message',
      required=True,
      help=Help.param_competition_message)
  parser_competitions_submit_optional.add_argument(
      '-v', '--version', dest='version', help=Help.param_code_version)
  parser_competitions_submit_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_competitions_submit._action_groups.append(
      parser_competitions_submit_optional)
  parser_competitions_submit.set_defaults(func=api.competition_submit_cli)

  # Competitions list submissions
  parser_competitions_submissions = subparsers_competitions.add_parser(
      'submissions',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_competitions_submissions)
  parser_competitions_submissions_optional = parser_competitions_submissions._action_groups.pop(
  )
  parser_competitions_submissions_optional.add_argument(
      'competition', nargs='?', default=None, help=Help.param_competition)
  parser_competitions_submissions_optional.add_argument(
      '-c',
      '--competition',
      dest='competition_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_competitions_submissions_optional.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_competitions_submissions_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_competitions_submissions._action_groups.append(
      parser_competitions_submissions_optional)
  parser_competitions_submissions.set_defaults(
      func=api.competition_submissions_cli)

  # Competitions leaderboard
  parser_competitions_leaderboard = subparsers_competitions.add_parser(
      'leaderboard',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_competitions_leaderboard)
  parser_competitions_leaderboard_optional = parser_competitions_leaderboard._action_groups.pop(
  )
  parser_competitions_leaderboard_optional.add_argument(
      'competition', nargs='?', default=None, help=Help.param_competition)
  parser_competitions_leaderboard_optional.add_argument(
      '-c',
      '--competition',
      dest='competition_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_competitions_leaderboard_optional.add_argument(
      '-s',
      '--show',
      dest='view',
      action='store_true',
      help=Help.param_competition_leaderboard_view)
  parser_competitions_leaderboard_optional.add_argument(
      '-d',
      '--download',
      dest='download',
      action='store_true',
      help=Help.param_competition_leaderboard_download)
  parser_competitions_leaderboard_optional.add_argument(
      '-p', '--path', dest='path', help=Help.param_downfolder)
  parser_competitions_leaderboard_optional.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_competitions_leaderboard_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_competitions_leaderboard._action_groups.append(
      parser_competitions_leaderboard_optional)
  parser_competitions_leaderboard.set_defaults(
      func=api.competition_leaderboard_cli)


def parse_datasets(subparsers):
  if six.PY2:
    parser_datasets = subparsers.add_parser(
        'datasets',
        formatter_class=argparse.RawTextHelpFormatter,
        help=Help.group_datasets)
  else:
    parser_datasets = subparsers.add_parser(
        'datasets',
        formatter_class=argparse.RawTextHelpFormatter,
        help=Help.group_datasets,
        aliases=['d'])
  subparsers_datasets = parser_datasets.add_subparsers(
      title='commands', dest='command')
  subparsers_datasets.required = True
  subparsers_datasets.choices = Help.datasets_choices

  # Datasets list
  parser_datasets_list = subparsers_datasets.add_parser(
      'list',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_datasets_list)
  parser_datasets_list_optional = parser_datasets_list._action_groups.pop()
  parser_datasets_list.add_argument(
      '--sort-by',
      dest='sort_by',
      required=False,
      help=Help.param_dataset_sort_by)
  parser_datasets_list.add_argument(
      '--size', dest='size', type=int, required=False, help=Help.param_dataset_size)
  parser_datasets_list.add_argument(
      '--file-type',
      dest='file_type',
      required=False,
      help=Help.param_dataset_file_type)
  parser_datasets_list.add_argument(
      '--license',
      dest='license_name',
      required=False,
      help=Help.param_dataset_license)
  parser_datasets_list.add_argument(
      '--tags', dest='tag_ids', required=False, help=Help.param_dataset_tags)
  parser_datasets_list.add_argument(
      '-s', '--search', dest='search', required=False, help=Help.param_search)
  parser_datasets_list.add_argument(
      '-m', '--mine', dest='mine', action='store_true', help=Help.param_mine)
  parser_datasets_list.add_argument(
      '--user', dest='user', required=False, help=Help.param_dataset_user)
  parser_datasets_list.add_argument(
      '-p',
      '--page',
      dest='page',
      default=1,
      type=int,
      required=False,
      help=Help.param_page)
  parser_datasets_list.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_datasets_list.add_argument(
      '--max-size',
      dest='max_size',
      required=False,
      type=int,
      help=Help.param_dataset_maxsize)
  parser_datasets_list.add_argument(
      '--min-size',
      dest='min_size',
      required=False,
      type=int,
      help=Help.param_dataset_minsize)
  parser_datasets_list._action_groups.append(parser_datasets_list_optional)
  parser_datasets_list.set_defaults(func=api.dataset_list_cli)

  # Datasets file list
  parser_datasets_files = subparsers_datasets.add_parser(
      'files',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_datasets_files)
  parser_datasets_files_optional = parser_datasets_files._action_groups.pop()
  parser_datasets_files_optional.add_argument(
      'dataset', nargs='?', default=None, help=Help.param_dataset)
  parser_datasets_files_optional.add_argument(
      '-d',
      '--dataset',
      dest='dataset_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_datasets_files_optional.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_datasets_files_optional.add_argument(
      '--page-token',
      dest='page_token',
      required=False,
      help=Help.param_page_token)
  parser_datasets_files_optional.add_argument(
      '--page-size',
      dest='page_size',
      required=False,
      default=20,
      type=int,
      help=Help.param_page_size)
  parser_datasets_files._action_groups.append(parser_datasets_files_optional)
  parser_datasets_files.set_defaults(func=api.dataset_list_files_cli)

  # Datasets download
  parser_datasets_download = subparsers_datasets.add_parser(
      'download',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_datasets_download)
  parser_datasets_download_optional = parser_datasets_download._action_groups.pop(
  )
  parser_datasets_download_optional.add_argument(
      'dataset', nargs='?', default=None, help=Help.param_dataset)
  parser_datasets_download_optional.add_argument(
      '-d',
      '--dataset',
      dest='dataset_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_datasets_download_optional.add_argument(
      '-f',
      '--file',
      dest='file_name',
      required=False,
      help=Help.param_dataset_file)
  parser_datasets_download_optional.add_argument(
      '-p', '--path', dest='path', required=False, help=Help.param_downfolder)
  parser_datasets_download_optional.add_argument(
      '-w',
      '--wp',
      dest='path',
      action='store_const',
      const='.',
      required=False,
      help=Help.param_wp)
  parser_datasets_download_optional.add_argument(
      '--unzip', dest='unzip', action='store_true', help=Help.param_unzip)
  parser_datasets_download_optional.add_argument(
      '-o', '--force', dest='force', action='store_true', help=Help.param_force)
  parser_datasets_download_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_datasets_download._action_groups.append(
      parser_datasets_download_optional)
  parser_datasets_download.set_defaults(func=api.dataset_download_cli)

  # Datasets create
  parser_datasets_create = subparsers_datasets.add_parser(
      'create',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_datasets_new)
  parser_datasets_create_optional = parser_datasets_create._action_groups.pop()
  parser_datasets_create_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_dataset_upfile)
  parser_datasets_create_optional.add_argument(
      '-u',
      '--public',
      dest='public',
      action='store_true',
      help=Help.param_public)
  parser_datasets_create_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_datasets_create_optional.add_argument(
      '-t',
      '--keep-tabular',
      dest='convert_to_csv',
      action='store_false',
      help=Help.param_keep_tabular)
  parser_datasets_create_optional.add_argument(
      '-r',
      '--dir-mode',
      dest='dir_mode',
      choices=['skip', 'zip', 'tar'],
      default='skip',
      help=Help.param_dir_mode)
  parser_datasets_create._action_groups.append(parser_datasets_create_optional)
  parser_datasets_create.set_defaults(func=api.dataset_create_new_cli)

  # Datasets update
  parser_datasets_version = subparsers_datasets.add_parser(
      'version',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_datasets_new_version)
  parser_datasets_version_optional = parser_datasets_version._action_groups.pop(
  )
  parser_datasets_version_required = parser_datasets_version.add_argument_group(
      'required arguments')
  parser_datasets_version_required.add_argument(
      '-m',
      '--message',
      dest='version_notes',
      required=True,
      help=Help.param_dataset_version_notes)
  parser_datasets_version_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_dataset_upfile)
  parser_datasets_version_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_datasets_version_optional.add_argument(
      '-t',
      '--keep-tabular',
      dest='convert_to_csv',
      action='store_false',
      help=Help.param_keep_tabular)
  parser_datasets_version_optional.add_argument(
      '-r',
      '--dir-mode',
      dest='dir_mode',
      choices=['skip', 'zip', 'tar'],
      default='skip',
      help=Help.param_dir_mode)
  parser_datasets_version_optional.add_argument(
      '-d',
      '--delete-old-versions',
      dest='delete_old_versions',
      action='store_true',
      help=Help.param_delete_old_version)
  parser_datasets_version._action_groups.append(
      parser_datasets_version_optional)
  parser_datasets_version.set_defaults(func=api.dataset_create_version_cli)

  # Datasets init
  parser_datasets_init = subparsers_datasets.add_parser(
      'init',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_datasets_init)
  parser_datasets_init_optional = parser_datasets_init._action_groups.pop()
  parser_datasets_init_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_dataset_upfile)
  parser_datasets_init._action_groups.append(parser_datasets_init_optional)
  parser_datasets_init.set_defaults(func=api.dataset_initialize_cli)

  # Datasets metadata
  parser_datasets_metadata = subparsers_datasets.add_parser(
      'metadata',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_datasets_metadata)
  parser_datasets_metadata_optional = parser_datasets_metadata._action_groups.pop(
  )
  parser_datasets_metadata_optional.add_argument(
      'dataset', nargs='?', default=None, help=Help.param_dataset)
  parser_datasets_metadata_optional.add_argument(
      '-d',
      '--dataset',
      dest='dataset_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_datasets_metadata_optional.add_argument(
      '--update',
      dest='update',
      action='store_true',
      help=Help.param_dataset_metadata_update)
  parser_datasets_metadata_optional.add_argument(
      '-p', '--path', dest='path', help=Help.param_dataset_metadata_dir)
  parser_datasets_metadata._action_groups.append(
      parser_datasets_metadata_optional)
  parser_datasets_metadata.set_defaults(func=api.dataset_metadata_cli)

  # Datasets status
  parser_datasets_status = subparsers_datasets.add_parser(
      'status',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_datasets_status)
  parser_datasets_status_optional = parser_datasets_status._action_groups.pop()
  parser_datasets_status_optional.add_argument(
      'dataset', nargs='?', default=None, help=Help.param_dataset)
  parser_datasets_status_optional.add_argument(
      '-d',
      '--dataset',
      dest='dataset_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_datasets_status._action_groups.append(parser_datasets_status_optional)
  parser_datasets_status.set_defaults(func=api.dataset_status_cli)


def parse_kernels(subparsers):
  if six.PY2:
    parser_kernels = subparsers.add_parser(
        'kernels',
        formatter_class=argparse.RawTextHelpFormatter,
        help=Help.group_kernels)
  else:
    parser_kernels = subparsers.add_parser(
        'kernels',
        formatter_class=argparse.RawTextHelpFormatter,
        help=Help.group_kernels,
        aliases=['k'])
  subparsers_kernels = parser_kernels.add_subparsers(
      title='commands', dest='command')
  subparsers_kernels.required = True
  subparsers_kernels.choices = Help.kernels_choices

  # Kernels list/search
  parser_kernels_list = subparsers_kernels.add_parser(
      'list',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_kernels_list)
  parser_kernels_list_optional = parser_kernels_list._action_groups.pop()
  parser_kernels_list_optional.add_argument(
      '-m', '--mine', dest='mine', action='store_true', help=Help.param_mine)
  parser_kernels_list_optional.add_argument(
      '-p', '--page', dest='page', default=1, type=int, help=Help.param_page)
  parser_kernels_list_optional.add_argument(
      '--page-size', dest='page_size', default=20, type=int, help=Help.param_page_size)
  parser_kernels_list_optional.add_argument(
      '-s', '--search', dest='search', help=Help.param_search)
  parser_kernels_list_optional.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_kernels_list_optional.add_argument(
      '--parent', dest='parent', required=False, help=Help.param_kernel_parent)
  parser_kernels_list_optional.add_argument(
      '--competition',
      dest='competition',
      required=False,
      help=Help.param_kernel_competition)
  parser_kernels_list_optional.add_argument(
      '--dataset',
      dest='dataset',
      required=False,
      help=Help.param_kernel_dataset)
  parser_kernels_list_optional.add_argument(
      '--user', dest='user', required=False, help=Help.param_kernel_user)
  parser_kernels_list_optional.add_argument(
      '--language',
      dest='language',
      required=False,
      help=Help.param_kernel_language)
  parser_kernels_list_optional.add_argument(
      '--kernel-type',
      dest='kernel_type',
      required=False,
      help=Help.param_kernel_type)
  parser_kernels_list_optional.add_argument(
      '--output-type',
      dest='output_type',
      required=False,
      help=Help.param_kernel_output_type)
  parser_kernels_list_optional.add_argument(
      '--sort-by',
      dest='sort_by',
      required=False,
      help=Help.param_kernel_sort_by)
  parser_kernels_list._action_groups.append(parser_kernels_list_optional)
  parser_kernels_list.set_defaults(func=api.kernels_list_cli)

  # Kernels file list
  parser_kernels_files = subparsers_kernels.add_parser(
      'files',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_kernels_files)
  parser_kernels_files_optional = parser_kernels_files._action_groups.pop()
  parser_kernels_files_optional.add_argument(
      'kernel', nargs='?', default=None, help=Help.param_kernel)
  parser_kernels_files_optional.add_argument(
      '-k',
      '--kernel',
      dest='kernel_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_kernels_files_optional.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_kernels_files_optional.add_argument(
      '--page-token', dest='page_token', help=Help.param_page_token)
  parser_kernels_files_optional.add_argument(
      '--page-size', dest='page_size', default=20, type=int, help=Help.param_page_size)
  parser_kernels_files._action_groups.append(parser_kernels_files_optional)
  parser_kernels_files.set_defaults(func=api.kernels_list_files_cli)

  # Kernels init
  parser_kernels_init = subparsers_kernels.add_parser(
      'init',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_kernels_init)
  parser_kernels_init_optional = parser_kernels_init._action_groups.pop()
  parser_kernels_init_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_kernel_upfile)
  parser_kernels_init._action_groups.append(parser_kernels_init_optional)
  parser_kernels_init.set_defaults(func=api.kernels_initialize_cli)

  # Kernels push
  parser_kernels_push = subparsers_kernels.add_parser(
      'push',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_kernels_push)
  parser_kernels_push_optional = parser_kernels_push._action_groups.pop()
  parser_kernels_push_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_kernel_upfile)
  parser_kernels_push_optional.add_argument(
      '-t',
      '--timeout',
      type=int,
      dest='timeout',
      help=Help.param_kernel_timeout)
  parser_kernels_push._action_groups.append(parser_kernels_push_optional)
  parser_kernels_push.set_defaults(func=api.kernels_push_cli)

  # Kernels pull
  parser_kernels_pull = subparsers_kernels.add_parser(
      'pull',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_kernels_pull)
  parser_kernels_pull_optional = parser_kernels_pull._action_groups.pop()
  parser_kernels_pull_optional.add_argument(
      'kernel', nargs='?', default=None, help=Help.param_kernel)
  parser_kernels_pull_optional.add_argument(
      '-k', '--kernel', dest='kernel', required=False, help=argparse.SUPPRESS)
  parser_kernels_pull_optional.add_argument(
      '-p', '--path', dest='path', required=False, help=Help.param_downfolder)
  parser_kernels_pull_optional.add_argument(
      '-w',
      '--wp',
      dest='path',
      action='store_const',
      const='.',
      required=False,
      help=Help.param_wp)
  parser_kernels_pull_optional.add_argument(
      '-m',
      '--metadata',
      dest='metadata',
      action='store_true',
      help=Help.param_kernel_pull_metadata)
  parser_kernels_pull._action_groups.append(parser_kernels_pull_optional)
  parser_kernels_pull.set_defaults(func=api.kernels_pull_cli)

  # Kernels output
  parser_kernels_output = subparsers_kernels.add_parser(
      'output',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_kernels_output)
  parser_kernels_output_optional = parser_kernels_output._action_groups.pop()
  parser_kernels_output_optional.add_argument(
      'kernel', nargs='?', default=None, help=Help.param_kernel)
  parser_kernels_output_optional.add_argument(
      '-k',
      '--kernel',
      dest='kernel_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_kernels_output_optional.add_argument(
      '-p', '--path', dest='path', required=False, help=Help.param_downfolder)
  parser_kernels_output_optional.add_argument(
      '-w',
      '--wp',
      dest='path',
      action='store_const',
      const='.',
      required=False,
      help=Help.param_wp)
  parser_kernels_output_optional.add_argument(
      '-o',
      '--force',
      dest='force',
      action='store_true',
      required=False,
      help=Help.param_force)
  parser_kernels_output_optional.add_argument(
      '-q',
      '--quiet',
      dest='quiet',
      action='store_true',
      required=False,
      help=Help.param_quiet)
  parser_kernels_output._action_groups.append(parser_kernels_output_optional)
  parser_kernels_output.set_defaults(func=api.kernels_output_cli)

  # Kernels status
  parser_kernels_status = subparsers_kernels.add_parser(
      'status',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_kernels_status)
  parser_kernels_status_optional = parser_kernels_status._action_groups.pop()
  parser_kernels_status_optional.add_argument(
      'kernel', nargs='?', default=None, help=Help.param_kernel)
  parser_kernels_status_optional.add_argument(
      '-k',
      '--kernel',
      dest='kernel_opt',
      required=False,
      help=argparse.SUPPRESS)
  parser_kernels_status._action_groups.append(parser_kernels_status_optional)
  parser_kernels_status.set_defaults(func=api.kernels_status_cli)


def parse_models(subparsers):
  parser_models = subparsers.add_parser(
      'models',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.group_models,
      aliases=['m'])

  subparsers_models = parser_models.add_subparsers(
      title='commands', dest='command')
  subparsers_models.required = True
  subparsers_models.choices = Help.models_choices

  # Models Instances.
  parse_model_instances(subparsers_models)

  # Models get
  parser_models_get = subparsers_models.add_parser(
      'get',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_models_get)
  parser_models_get_optional = parser_models_get._action_groups.pop()
  parser_models_get_optional.add_argument('model', help=Help.param_model)
  parser_models_get_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_model_downfile)
  parser_models_get._action_groups.append(parser_models_get_optional)
  parser_models_get.set_defaults(func=api.model_get_cli)

  # Models list
  parser_models_list = subparsers_models.add_parser(
      'list',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_models_list)
  parser_models_list_optional = parser_models_list._action_groups.pop()
  parser_models_list.add_argument(
      '--sort-by',
      dest='sort_by',
      required=False,
      help=Help.param_model_sort_by)
  parser_models_list.add_argument(
      '-s', '--search', dest='search', required=False, help=Help.param_search)
  parser_models_list.add_argument(
      '--owner', dest='owner', required=False, help=Help.param_model_owner)
  parser_models_list.add_argument(
      '--page-size', dest='page_size', default=20, type=int, help=Help.param_page_size)
  parser_models_list.add_argument(
      '--page-token',
      dest='page_token',
      required=False,
      help=Help.param_page_token)
  parser_models_list.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_models_list._action_groups.append(parser_models_list_optional)
  parser_models_list.set_defaults(func=api.model_list_cli)

  # Models init
  parser_models_init = subparsers_models.add_parser(
      'init',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_models_init)
  parser_models_init_optional = parser_models_init._action_groups.pop()
  parser_models_init_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_model_upfile)
  parser_models_init._action_groups.append(parser_models_init_optional)
  parser_models_init.set_defaults(func=api.model_initialize_cli)

  # Models create
  parser_models_create = subparsers_models.add_parser(
      'create',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_models_new)
  parser_models_create_optional = parser_models_create._action_groups.pop()
  parser_models_create_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_model_upfile)
  parser_models_create._action_groups.append(parser_models_create_optional)
  parser_models_create.set_defaults(func=api.model_create_new_cli)

  # Models delete
  parser_models_delete = subparsers_models.add_parser(
      'delete',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_models_delete)
  parser_models_delete_optional = parser_models_delete._action_groups.pop()
  parser_models_delete_optional.add_argument('model', help=Help.param_model)
  parser_models_delete_optional.add_argument(
      '-y', '--yes', dest='yes', action='store_true', help=Help.param_yes)
  parser_models_delete._action_groups.append(parser_models_delete_optional)
  parser_models_delete.set_defaults(func=api.model_delete_cli)

  # Models update
  parser_models_update = subparsers_models.add_parser(
      'update',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_models_update)
  parser_models_update_optional = parser_models_update._action_groups.pop()
  parser_models_update_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_model_upfile)
  parser_models_update._action_groups.append(parser_models_update_optional)
  parser_models_update.set_defaults(func=api.model_update_cli)


def parse_model_instances(subparsers):
  parser_model_instances = subparsers.add_parser(
      'instances',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.group_model_instances,
      aliases=['mi'])

  subparsers_model_instances = parser_model_instances.add_subparsers(
      title='commands', dest='command')
  subparsers_model_instances.required = True
  subparsers_model_instances.choices = Help.model_instances_choices

  # Models Instances Versions.
  parse_model_instance_versions(subparsers_model_instances)

  # Models Instances get
  parser_model_instance_get = subparsers_model_instances.add_parser(
      'get',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instances_get)
  parser_model_instance_get_optional = parser_model_instance_get._action_groups.pop(
  )
  parser_model_instance_get_optional.add_argument(
      'model_instance', help=Help.param_model_instance)
  parser_model_instance_get_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_model_instance_downfile)
  parser_model_instance_get._action_groups.append(
      parser_model_instance_get_optional)
  parser_model_instance_get.set_defaults(func=api.model_instance_get_cli)

  # Model Instances init
  parser_model_instances_init = subparsers_model_instances.add_parser(
      'init',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instances_init)
  parser_model_instances_init_optional = parser_model_instances_init._action_groups.pop(
  )
  parser_model_instances_init_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_model_instance_upfile)
  parser_model_instances_init._action_groups.append(
      parser_model_instances_init_optional)
  parser_model_instances_init.set_defaults(
      func=api.model_instance_initialize_cli)

  # Model Instances create
  parser_model_instances_create = subparsers_model_instances.add_parser(
      'create',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instances_new)
  parser_model_instances_create_optional = parser_model_instances_create._action_groups.pop(
  )
  parser_model_instances_create_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_model_instance_upfile)
  parser_model_instances_create_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_model_instances_create_optional.add_argument(
      '-r',
      '--dir-mode',
      dest='dir_mode',
      choices=['skip', 'zip', 'tar'],
      default='skip',
      help=Help.param_dir_mode)
  parser_model_instances_create._action_groups.append(
      parser_model_instances_create_optional)
  parser_model_instances_create.set_defaults(func=api.model_instance_create_cli)

  # Model Instances files
  parser_model_instances_files = subparsers_model_instances.add_parser(
      'files',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instances_files)
  parser_model_instances_files_optional = parser_model_instances_files._action_groups.pop(
  )
  parser_model_instances_files_optional.add_argument(
      'model_instance', help=Help.param_model_instance)
  parser_model_instances_files_optional.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_model_instances_files_optional.add_argument(
      '--page-size', dest='page_size', default=20, type=int, help=Help.param_page_size)
  parser_model_instances_files_optional.add_argument(
      '--page-token',
      dest='page_token',
      required=False,
      help=Help.param_page_token)
  parser_model_instances_files._action_groups.append(
      parser_model_instances_files_optional)
  parser_model_instances_files.set_defaults(func=api.model_instance_files_cli)

  # Models Instances delete
  parser_model_instances_delete = subparsers_model_instances.add_parser(
      'delete',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instances_delete)
  parser_model_instances_delete_optional = parser_model_instances_delete._action_groups.pop(
  )
  parser_model_instances_delete_optional.add_argument(
      'model_instance', help=Help.param_model_instance)
  parser_model_instances_delete_optional.add_argument(
      '-y', '--yes', dest='yes', action='store_true', help=Help.param_yes)
  parser_model_instances_delete._action_groups.append(
      parser_model_instances_delete_optional)
  parser_model_instances_delete.set_defaults(func=api.model_instance_delete_cli)

  # Models Instances update
  parser_model_instances_update = subparsers_model_instances.add_parser(
      'update',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instances_update)
  parser_model_instances_update_optional = parser_model_instances_update._action_groups.pop(
  )
  parser_model_instances_update_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_model_instance_upfile)
  parser_model_instances_update._action_groups.append(
      parser_model_instances_update_optional)
  parser_model_instances_update.set_defaults(func=api.model_instance_update_cli)


def parse_model_instance_versions(subparsers):
  parser_model_instance_versions = subparsers.add_parser(
      'versions',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.group_model_instance_versions,
      aliases=['miv'])

  subparsers_model_intance_versions = parser_model_instance_versions.add_subparsers(
      title='commands', dest='command')
  subparsers_model_intance_versions.required = True
  subparsers_model_intance_versions.choices = Help.model_instance_versions_choices

  # Model Instance Versions create
  parser_model_instance_versions_create = subparsers_model_intance_versions.add_parser(
      'create',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instance_versions_new)
  parser_model_instance_versions_create_optional = parser_model_instance_versions_create._action_groups.pop(
  )
  parser_model_instance_versions_create_optional.add_argument(
      'model_instance', help=Help.param_model_instance)
  parser_model_instance_versions_create_optional.add_argument(
      '-p',
      '--path',
      dest='folder',
      required=False,
      help=Help.param_model_instance_version_upfile)
  parser_model_instance_versions_create_optional.add_argument(
      '-n',
      '--version-notes',
      dest='version_notes',
      required=False,
      help=Help.param_model_instance_version_notes)
  parser_model_instance_versions_create_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_model_instance_versions_create_optional.add_argument(
      '-r',
      '--dir-mode',
      dest='dir_mode',
      choices=['skip', 'zip', 'tar'],
      default='skip',
      help=Help.param_dir_mode)
  parser_model_instance_versions_create._action_groups.append(
      parser_model_instance_versions_create_optional)
  parser_model_instance_versions_create.set_defaults(
      func=api.model_instance_version_create_cli)

  # Models Instance Versions download
  parser_model_instance_versions_download = subparsers_model_intance_versions.add_parser(
      'download',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instance_versions_download)
  parser_model_instance_versions_download_optional = parser_model_instance_versions_download._action_groups.pop(
  )
  parser_model_instance_versions_download_optional.add_argument(
      'model_instance_version', help=Help.param_model_instance_version)
  parser_model_instance_versions_download_optional.add_argument(
      '-p', '--path', dest='path', required=False, help=Help.param_downfolder)
  parser_model_instance_versions_download_optional.add_argument(
      '--untar', dest='untar', action='store_true', help=Help.param_untar)
  parser_model_instance_versions_download_optional.add_argument(
      '-f', '--force', dest='force', action='store_true', help=Help.param_force)
  parser_model_instance_versions_download_optional.add_argument(
      '-q', '--quiet', dest='quiet', action='store_true', help=Help.param_quiet)
  parser_model_instance_versions_download._action_groups.append(
      parser_model_instance_versions_download_optional)
  parser_model_instance_versions_download.set_defaults(
      func=api.model_instance_version_download_cli)

  # Models Instance Versions files
  parser_model_instance_versions_files = subparsers_model_intance_versions.add_parser(
      'files',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instance_versions_files)
  parser_model_instance_versions_files_optional = parser_model_instance_versions_files._action_groups.pop(
  )
  parser_model_instance_versions_files_optional.add_argument(
      'model_instance_version', help=Help.param_model_instance_version)
  parser_model_instance_versions_files_optional.add_argument(
      '-v',
      '--csv',
      dest='csv_display',
      action='store_true',
      help=Help.param_csv)
  parser_model_instance_versions_files_optional.add_argument(
      '--page-size', dest='page_size', default=20, type=int, help=Help.param_page_size)
  parser_model_instance_versions_files_optional.add_argument(
      '--page-token',
      dest='page_token',
      required=False,
      help=Help.param_page_token)
  parser_model_instance_versions_files._action_groups.append(
      parser_model_instance_versions_files_optional)
  parser_model_instance_versions_files.set_defaults(
      func=api.model_instance_version_files_cli)

  # Models Instance Versions delete
  parser_model_instance_versions_delete = subparsers_model_intance_versions.add_parser(
      'delete',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_model_instance_versions_delete)
  parser_model_instance_versions_delete_optional = parser_model_instance_versions_delete._action_groups.pop(
  )
  parser_model_instance_versions_delete_optional.add_argument(
      'model_instance_version', help=Help.param_model_instance_version)
  parser_model_instance_versions_delete_optional.add_argument(
      '-y', '--yes', dest='yes', action='store_true', help=Help.param_yes)
  parser_model_instance_versions_delete._action_groups.append(
      parser_model_instance_versions_delete_optional)
  parser_model_instance_versions_delete.set_defaults(
      func=api.model_instance_version_delete_cli)


def parse_files(subparsers):
  parser_files = subparsers.add_parser(
      'files',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.group_files,
      aliases=['f'])

  subparsers_files = parser_files.add_subparsers(
      title='commands', dest='command')
  subparsers_files.required = True
  subparsers_files.choices = Help.files_choices

  # Files upload
  parser_files_upload = subparsers_files.add_parser(
      'upload',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_files_upload,
      aliases=['u'])
  parser_files_upload_optional = parser_files_upload._action_groups.pop()
  parser_files_upload_optional.add_argument(
      '-i',
      '--inbox-path',
      dest='inbox_path',
      required=False,
      default='',
      help=Help.param_files_upload_inbox_path)
  parser_files_upload_optional.add_argument(
      'local_paths',
      metavar='local-path',
      nargs='+',
      help=Help.param_files_upload_local_paths)
  parser_files_upload_optional.add_argument(
      '--no-resume',
      dest='no_resume',
      action='store_true',
      required=False,
      default=False,
      help=Help.param_files_upload_no_resume)
  parser_files_upload_optional.add_argument(
      '--no-compress',
      dest='no_compress',
      action='store_true',
      required=False,
      default=False,
      help=Help.param_files_upload_no_compress)
  parser_files_upload._action_groups.append(parser_files_upload_optional)
  parser_files_upload.set_defaults(func=api.files_upload_cli)


def parse_config(subparsers):
  parser_config = subparsers.add_parser(
      'config',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.group_config)
  subparsers_config = parser_config.add_subparsers(
      title='commands', dest='command')
  subparsers_config.required = True
  subparsers_config.choices = Help.config_choices

  parser_config_view = subparsers_config.add_parser(
      'view',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_config_view)
  parser_config_view.set_defaults(func=api.print_config_values)

  parser_config_set = subparsers_config.add_parser(
      'set',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_config_set)
  parser_config_set._action_groups.pop()
  parser_config_set_required = parser_config_set.add_argument_group(
      'required arguments')
  parser_config_set_required.add_argument(
      '-n', '--name', dest='name', required=True, help=Help.param_config_name)
  parser_config_set_required.add_argument(
      '-v',
      '--value',
      dest='value',
      required=True,
      help=Help.param_config_value)
  parser_config_set.set_defaults(func=api.set_config_value)

  parser_config_unset = subparsers_config.add_parser(
      'unset',
      formatter_class=argparse.RawTextHelpFormatter,
      help=Help.command_config_unset)
  parser_config_unset._action_groups.pop()
  parser_config_unset_required = parser_config_unset.add_argument_group(
      'required arguments')
  parser_config_unset_required.add_argument(
      '-n', '--name', dest='name', required=True, help=Help.param_config_name)
  parser_config_unset.set_defaults(func=api.unset_config_value)


class Help(object):
  kaggle_choices = [
      'competitions', 'c', 'datasets', 'd', 'kernels', 'k', 'models', 'm',
      'files', 'f', 'config'
  ]
  competitions_choices = [
      'list', 'files', 'download', 'submit', 'submissions', 'leaderboard'
  ]
  datasets_choices = [
      'list', 'files', 'download', 'create', 'version', 'init', 'metadata',
      'status'
  ]
  kernels_choices = [
      'list', 'files', 'init', 'push', 'pull', 'output', 'status'
  ]
  models_choices = [
      'instances', 'get', 'list', 'init', 'create', 'delete', 'update'
  ]
  model_instances_choices = [
      'versions', 'get', 'files', 'init', 'create', 'delete', 'update'
  ]
  model_instance_versions_choices = [
      'init', 'create', 'download', 'delete', 'files'
  ]
  files_choices = ['upload']
  config_choices = ['view', 'set', 'unset']

  kaggle = 'Use one of:\ncompetitions {' + ', '.join(
      competitions_choices) + '}\ndatasets {' + ', '.join(
          datasets_choices) + '}\nkernels {' + ', '.join(
              kernels_choices) + '}\nmodels {' + ', '.join(
                  models_choices) + '}\nmodels instances {' + ', '.join(
                      model_instances_choices
                  ) + '}\nmodels instances versions {' + ', '.join(
                      model_instance_versions_choices
                  ) + '}\nconfig {' + ', '.join(config_choices) + '}'

  group_competitions = 'Commands related to Kaggle competitions'
  group_datasets = 'Commands related to Kaggle datasets'
  group_kernels = 'Commands related to Kaggle kernels'
  group_models = 'Commands related to Kaggle models'
  group_model_instances = 'Commands related to Kaggle model instances'
  group_model_instance_versions = 'Commands related to Kaggle model instance versions'
  group_files = 'Commands related files'
  group_config = 'Configuration settings'

  # Competitions commands
  command_competitions_list = 'List available competitions'
  command_competitions_files = 'List competition files'
  command_competitions_download = 'Download competition files'
  command_competitions_submit = 'Make a new competition submission'
  command_competitions_submissions = 'Show your competition submissions'
  command_competitions_leaderboard = 'Get competition leaderboard information'

  # Datasets commands
  command_datasets_list = 'List available datasets'
  command_datasets_files = 'List dataset files'
  command_datasets_download = 'Download dataset files'
  command_datasets_new = 'Create a new dataset'
  command_datasets_new_version = 'Create a new dataset version'
  command_datasets_init = 'Initialize metadata file for dataset creation'
  command_datasets_metadata = 'Download metadata about a dataset'
  command_datasets_status = 'Get the creation status for a dataset'

  # Kernels commands
  command_kernels_list = (
      'List available kernels. By default, shows 20 results sorted by '
      'hotness')
  command_kernels_files = 'List kernel output files'
  command_kernels_init = 'Initialize metadata file for a kernel'
  command_kernels_push = 'Push new code to a kernel and run the kernel'
  command_kernels_pull = 'Pull down code from a kernel'
  command_kernels_output = 'Get data output from the latest kernel run'
  command_kernels_status = 'Display the status of the latest kernel run'

  # Models commands
  command_models_files = 'List model files'
  command_models_get = 'Get a model'
  command_models_list = 'List models'
  command_models_init = 'Initialize metadata file for model creation'
  command_models_new = 'Create a new model'
  command_models_delete = 'Delete a model'
  command_models_update = 'Update a model'

  # Files commands
  command_files_upload = 'Upload files'

  # Config commands
  command_config_path = (
      'Set folder where competition or dataset files will be '
      'downloaded')
  command_config_proxy = 'Set proxy server'
  command_config_competition = 'Set default competition'
  command_config_view = 'View current config values'
  command_config_set = 'Set a configuration value'
  command_config_unset = 'Clear a configuration value'

  # General params
  param_downfolder = (
      'Folder where file(s) will be downloaded, defaults to current working '
      'directory')
  param_wp = 'Download files to current working path'
  param_proxy = 'Proxy for HTTP requests'
  param_quiet = (
      'Suppress printing information about the upload/download progress')
  param_public = 'Create publicly (default is private)'
  param_keep_tabular = (
      'Do not convert tabular files to CSV (default is to convert)')
  param_dir_mode = (
      'What to do with directories: "skip" - ignore; "zip" - compressed upload; "tar" - '
      'uncompressed upload')
  param_delete_old_version = 'Delete old versions of this dataset'
  param_force = ('Skip check whether local version of file is up to date, force'
                 ' file download')
  param_upfile = 'File for upload (full path), or the name of the output file produced by a kernel (for code competitions)'
  param_code_kernel = 'Name of kernel (notebook) to submit to a code competition'
  param_code_version = 'Version of kernel to submit to a code competition, e.g. "Version 1"'
  param_csv = 'Print results in CSV format (if not set print in table format)'
  param_page = 'Page number for results paging. Page size is 20 by default'
  # NOTE: Default and max page size are set by the mid-tier code.
  param_page_size = ('Number of items to show on a page. Default size is 20, '
                     'max is 200')
  param_page_token = 'Page token for results paging.'
  param_search = 'Term(s) to search for'
  param_mine = 'Display only my items'
  param_unzip = (
      'Unzip the downloaded file. Will delete the zip file when completed.')
  param_untar = (
      'Untar the downloaded file. Will delete the tar file when completed.')
  param_yes = (
      'Sets any confirmation values to "yes" automatically. Users will not be asked to confirm.'
  )

  # Competitions params
  param_competition = ('Competition URL suffix (use "kaggle competitions list" '
                       'to show options)\nIf empty, the default competition '
                       'will be used (use "kaggle config set competition")"')
  param_competition_nonempty = (
      'Competition URL suffix (use "kaggle competitions list" to show '
      'options)')
  param_competition_leaderboard_view = 'Show the top of the leaderboard'
  param_competition_leaderboard_download = 'Download entire leaderboard'
  param_competition_file = (
      'File name, all files downloaded if not provided\n(use "kaggle '
      'competitions files -c <competition>" to show options)')
  param_competition_message = 'Message describing this submission'
  param_competition_group = (
      'Search for competitions in a specific group. Default is \'general\'. '
      'Valid options are \'general\', \'entered\', and \'inClass\'')
  param_competition_category = (
      'Search for competitions of a specific category. Default is \'all\'. '
      'Valid options are \'all\', \'featured\', \'research\', '
      '\'recruitment\', \'gettingStarted\', \'masters\', and \'playground\'')
  param_competition_sort_by = (
      'Sort list results. Default is \'latestDeadline\'. Valid options are '
      '\'grouped\', \'prize\', \'earliestDeadline\', \'latestDeadline\', '
      '\'numberOfTeams\', and \'recentlyCreated\'')

  # Datasets params
  param_dataset = ('Dataset URL suffix in format <owner>/<dataset-name> (use '
                   '"kaggle datasets list" to show options)')
  param_dataset_file = ('File name, all files downloaded if not provided\n(use '
                        '"kaggle datasets files -d <dataset>" to show options)')
  param_dataset_version_notes = 'Message describing the new version'
  param_dataset_upfile = (
      'Folder for upload, containing data files and a '
      'special datasets-metadata.json file '
      '(https://github.com/Kaggle/kaggle-api/wiki/Dataset-Metadata). '
      'Defaults to current working directory')
  param_dataset_sort_by = (
      'Sort list results. Default is \'hottest\'. Valid options are '
      '\'hottest\', \'votes\', \'updated\', and \'active\'')
  param_dataset_size = (
      'DEPRECATED. Please use --max-size and --min-size to filter dataset sizes.'
  )
  param_dataset_file_type = (
      'Search for datasets with a specific file type. Default is \'all\'. '
      'Valid options are \'all\', \'csv\', \'sqlite\', \'json\', and '
      '\'bigQuery\'. Please note that bigQuery datasets cannot be downloaded')
  param_dataset_license = (
      'Search for datasets with a specific license. Default is \'all\'. '
      'Valid options are \'all\', \'cc\', \'gpl\', \'odb\', and \'other\'')
  param_dataset_tags = (
      'Search for datasets that have specific tags. Tag list should be '
      'comma separated')
  param_dataset_user = (
      'Find public datasets owned by a specific user or organization')
  param_dataset_metadata_dir = (
      'Location to download dataset metadata to. Defaults to current working '
      'directory')
  param_dataset_metadata_update = ('A flag to indicate whether the dataset'
                                   'metadata should be updated.')
  param_dataset_maxsize = 'Specify the maximum size of the dataset to return (bytes)'
  param_dataset_minsize = 'Specify the minimum size of the dataset to return (bytes)'

  # Kernels params
  param_kernel = (
      'Kernel URL suffix in format <owner>/<kernel-name> (use "kaggle '
      'kernels list" to show options)')
  param_kernel_init = (
      'Create a metadata file for an existing kernel URL suffix in format '
      '<owner>/<kernel-name> (use "kaggle kernels list" to show options)')
  param_kernel_upfile = (
      'Folder for upload, containing data files and a '
      'special kernel-metadata.json file '
      '(https://github.com/Kaggle/kaggle-api/wiki/Kernel-Metadata). '
      'Defaults to current working directory')
  param_kernel_parent = 'Find children of the specified parent kernel'
  param_kernel_competition = 'Find kernels for a given competition slug'
  param_kernel_dataset = ('Find kernels for a given dataset slug. Format is '
                          '{username/dataset-slug}')
  param_kernel_timeout = ('Limit the run time of a kernel to the given number '
                          'of seconds. The global maximum time will not be '
                          'exceeded.')
  param_kernel_user = 'Find kernels created by a given username'
  # TODO(b/129357583): Pull these from the same spot as the api impl
  param_kernel_language = (
      'Specify the language the kernel is written in. Default is \'all\'. '
      'Valid options are \'all\', \'python\', \'r\', \'sqlite\', and '
      '\'julia\'')
  param_kernel_type = ('Specify the type of kernel. Default is \'all\'. Valid '
                       'options are \'all\', \'script\', and \'notebook\'')
  param_kernel_output_type = ('Search for specific kernel output types. '
                              'Default is \'all\'.  Valid options are \'all\', '
                              '\'visualizations\', and \'data\'')
  param_kernel_sort_by = ('Sort list results. Default is \'hotness\'. Valid '
                          'options are \'hotness\', \'commentCount\', '
                          '\'dateCreated\', \'dateRun\', \'relevance\', '
                          '\'scoreAscending\', \'scoreDescending\', '
                          '\'viewCount\', and \'voteCount\'. \'relevance\' '
                          'is only applicable if a search term is specified.')
  param_kernel_pull_metadata = 'Generate metadata when pulling kernel'

  # Models params
  param_model = ('Model URL suffix in format <owner>/<model-name>')
  param_model_sort_by = (
      'Sort list results. Default is \'hotness\'. Valid options are '
      '\'hotness\', \'downloadCount\', \'voteCount\', \'notebookCount\' and \'createTime\''
  )
  param_model_owner = (
      'Find public models owned by a specific user or organization')
  param_model_downfile = (
      'Folder containing the special model-metadata.json file '
      '(https://github.com/Kaggle/kaggle-api/wiki/Model-Metadata).')
  param_model_upfile = (
      'Folder containing the special model-metadata.json file '
      '(https://github.com/Kaggle/kaggle-api/wiki/Model-Metadata). '
      'Defaults to current working directory')

  # Model Instances params
  param_model_instance = (
      'Model Instance URL suffix in format <owner>/<model-name>/<framework>/<instance-slug>'
  )
  command_model_instances_get = 'Get a model instance'
  command_model_instances_init = 'Initialize metadata file for model instance creation'
  command_model_instances_files = 'List files for the current version of a model instance'
  command_model_instances_new = 'Create a new model instance'
  param_model_instance_downfile = (
      'Folder for downloading the special model-instance-metadata.json file '
      '(https://github.com/Kaggle/kaggle-api/wiki/Model-Metadata#model-instance). '
  )
  param_model_instance_upfile = (
      'Folder for upload, containing data files and a '
      'special model-instance-metadata.json file '
      '(https://github.com/Kaggle/kaggle-api/wiki/Model-Metadata#model-instance). '
      'Defaults to current working directory')
  command_model_instances_delete = 'Delete a model instance'
  command_model_instances_update = 'Update a model instance'

  # Model Instance Versions params
  param_model_instance_version = (
      'Model Instance Version URL suffix in format <owner>/<model-name>/<framework>/<instance-slug>/<version-number>'
  )

  # Model Instance Versions params
  command_model_instance_versions_new = 'Create a new model instance version'
  param_model_instance_version_upfile = (
      'Folder for upload. Defaults to current working directory')
  command_model_instance_versions_delete = 'Delete a model instance version'
  command_model_instance_versions_download = 'Download model instance version files'
  command_model_instance_versions_files = 'List model instance version files'
  param_model_instance_version_notes = 'Version notes to record for the new model instance version'

  # Files params
  param_files_upload_inbox_path = 'Virtual path on the server where the uploaded files will be stored'
  param_files_upload_local_paths = (
      'List of local filesystem paths. Each path creates a separate file on the server. '
      'Directories are uploaded as zip archives by default (e.g., a directory called '
      '"data" will be uploaded as "data.zip")')
  param_files_upload_no_compress = 'Whether to compress directories (zip) or not (tar)'
  param_files_upload_no_resume = 'Whether to skip resumable uploads.'

  # Config params
  param_config_name = ('Name of the configuration parameter\n(one of '
                       'competition, path, proxy)')
  param_config_value = (('Value of the configuration parameter, valid values '
                         'depending on name\n- competition: ') +
                        param_competition_nonempty + '\n- path: ' +
                        param_downfolder + '\n- proxy: ' + param_proxy)


if __name__ == '__main__':
  main()



import binascii
import codecs
import json
import os
import urllib.parse
from io import BytesIO

import requests
from urllib3.fields import RequestField

from kagglesdk.kaggle_env import get_endpoint, get_env, KaggleEnv
from kagglesdk.kaggle_object import KaggleObject
from typing import Type

# TODO (http://b/354237483) Generate the client from the existing one.
# This was created from kaggle_api_client.py, prior to recent changes to
# auth handling. The new client requires KAGGLE_API_TOKEN, so it is not
# currently usable by the CLI.

# TODO: Extend kapigen to add a boolean to these requests indicating that they use forms.
REQUESTS_REQUIRING_FORMS = [
    'ApiUploadDatasetFileRequest',
    'ApiCreateSubmissionRequest',
    'ApiCreateCodeSubmissionRequest',
    'ApiStartSubmissionUploadRequest',
    'ApiUploadModelFileRequest',
]


def _headers_to_str(headers):
  return '\n'.join(f'{k}: {v}' for k, v in headers.items())


def _get_apikey_creds():
  apikey_filename = os.path.expanduser('~/.kaggle/kaggle.json')
  if not os.path.exists(apikey_filename):
    return None

  kaggle_json = None
  with open(apikey_filename) as apikey_file:
    kaggle_json = apikey_file.read()

  if not kaggle_json or not kaggle_json.strip():
    return None

  api_key_data = json.loads(kaggle_json)
  username = api_key_data['username']
  api_key = api_key_data['key']
  return username, api_key


def clean_data(data):
  if isinstance(data, dict):
    return {
        to_lower_camel_case(k): clean_data(v) for k, v in data.items() if v is not None
    }
  if isinstance(data, list):
    return [clean_data(v) for v in data if v is not None]
  if data is True:
    return 'true'
  if data is False:
    return 'false'
  return data


def find_words(source, left='{', right='}'):
  words = []
  split_str = source.split(left)

  for s in split_str[1:]:
    split_s = s.split(right)
    if len(split_s) > 1:
      words.append(split_s[0])

  return words


def to_camel_case(snake_str):
  return ''.join(x.capitalize() for x in snake_str.lower().split('_'))


def to_lower_camel_case(snake_str):
  # https://stackoverflow.com/questions/19053707/converting-snake-case-to-lower-camel-case-lowercamelcase
  # We capitalize the first letter of each component except the first one
  # with the 'capitalize' method and join them together.
  camel_string = to_camel_case(snake_str)
  return snake_str[0].lower() + camel_string[1:]


class KaggleHttpClient(object):
  _xsrf_cookie_name = 'XSRF-TOKEN'
  _csrf_cookie_name = 'CSRF-TOKEN'
  _xsrf_cookies = (_xsrf_cookie_name, _csrf_cookie_name)
  _xsrf_header_name = 'X-XSRF-TOKEN'

  def __init__(
      self,
      env: KaggleEnv = None,
      verbose: bool = False,
      renew_iap_token=None,
      username=None,
      password=None,
  ):
    self._env = env or get_env()
    self._signed_in = None
    self._endpoint = get_endpoint(self._env)
    self._verbose = verbose
    self._session = None
    self._username = username
    self._password = password

  def call(
      self,
      service_name: str,
      request_name: str,
      request: KaggleObject,
      response_type: Type[KaggleObject],
  ):
    self._init_session()
    http_request = self._prepare_request(service_name, request_name, request)

    # Merge environment settings into session
    settings = self._session.merge_environment_settings(http_request.url, {}, None, None, None)
    http_response = self._session.send(http_request, **settings)

    response = self._prepare_response(response_type, http_response)
    return response

  def _prepare_request(
      self, service_name: str, request_name: str, request: KaggleObject
  ):
    request_url = self._get_request_url(request)
    method = request.method()
    data = ''
    if method == 'GET':
      data = request.__class__.to_dict(request, ignore_defaults=False)
      if request.endpoint_path():
        words = find_words(request.endpoint_path())
        list(map(data.pop, [to_lower_camel_case(w) for w in words]))
        if len(data) == 0:
          data = None
      if data:
        request_url = f'{request_url}?{urllib.parse.urlencode(clean_data(data))}'
      data = ''
      self._session.headers.update(
          {
              'Accept': 'application/json',
              'Content-Type': 'text/plain',
          }
      )
    elif method == 'POST':
      data = request.to_field_map(request, ignore_defaults=True)
      if isinstance(data, dict):
        fields = request.body_fields()
        if fields is not None:
          if fields != '*':
            data = data[fields]
        data = clean_data(data)
        if self.requires_form(request):
          data, content_type = self.make_form(data)
        else:
          content_type = 'application/json'
          data = json.dumps(data)
        self._session.headers.update(
            {
                'Accept': 'application/json',
                'Content-Type': content_type,
            }
        )
    http_request = requests.Request(
        method=method,
        url=request_url,
        data=data,
        headers=self._session.headers,
        # cookies=self._get_xsrf_cookies(),
        auth=self._session.auth,
    )
    prepared_request = http_request.prepare()
    self._print_request(prepared_request)
    return prepared_request

  def _get_xsrf_cookies(self):
    cookies = requests.cookies.RequestsCookieJar()
    for cookie in self._session.cookies:
      if cookie.name in KaggleHttpClient._xsrf_cookies:
        cookies[cookie.name] = cookie.value
    return cookies

  def _prepare_response(self, response_type, http_response):
    self._print_response(http_response)
    http_response.raise_for_status()
    if 'application/json' in http_response.headers['Content-Type']:
      resp = http_response.json()
      if 'code' in resp and resp['code'] >= 400:
        raise requests.exceptions.HTTPError(resp['message'], response=http_response)
    if response_type is None:  # Method doesn't have a return type
      return None
    return response_type.prepare_from(http_response)

  def _print_request(self, request):
    if not self._verbose:
      return
    self._print('---------------------Request----------------------')
    self._print(
        f'{request.method} {request.url}\n{_headers_to_str(request.headers)}\n\n{request.body}'
    )
    self._print('--------------------------------------------------')

  def _print_response(self, response, body=True):
    if not self._verbose:
      return
    self._print('---------------------Response---------------------')
    self._print(f'{response.status_code}\n{_headers_to_str(response.headers)}')
    if body:
      self._print(f'\n{response.text}')
    self._print('--------------------------------------------------')

  def _print(self, message: str):
    if self._verbose:
      print(message)

  def __enter__(self):
    self._init_session()
    return self

  def __exit__(self, exc_type, exc_value, tb):
    if self._session is not None:
      self._session.close()

  def _init_session(self):
    if self._session is not None:
      return self._session

    self._session = requests.Session()
    self._session.headers.update(
        {
            'User-Agent': 'kaggle-api/v1.7.0',  # Was: V2
            'Content-Type': 'application/x-www-form-urlencoded',  # Was: /json
        }
    )

    iap_token = self._get_iap_token_if_required()
    if iap_token is not None:
      self._session.headers.update(
          {
              # https://cloud.google.com/iap/docs/authentication-howto#authenticating_from_proxy-authorization_header
              'Proxy-Authorization': f'Bearer {iap_token}',
          }
      )

    self._try_fill_auth()
    # self._fill_xsrf_token(iap_token)  # TODO Make this align with original handler.

  def _get_iap_token_if_required(self):
    if self._env not in (KaggleEnv.STAGING, KaggleEnv.ADMIN):
      return None
    iap_token = os.getenv('KAGGLE_IAP_TOKEN')
    if iap_token is None:
      raise Exception(f'Must set KAGGLE_IAP_TOKEN to access "{self._endpoint}"')
    return iap_token

  def _fill_xsrf_token(self, iap_token):
    initial_get_request = requests.Request(
        method='GET',
        url=self._endpoint,
        headers=self._session.headers,
        auth=self._session.auth,
    )
    prepared_request = initial_get_request.prepare()
    self._print_request(prepared_request)

    http_response = self._session.send(prepared_request)

    self._print_response(http_response, body=False)
    if iap_token is not None and http_response.status_code in (401, 403):
      raise requests.exceptions.HTTPError('IAP token invalid or expired')
    http_response.raise_for_status()

    self._session.headers.update(
        {
            KaggleHttpClient._xsrf_header_name: self._session.cookies[
                KaggleHttpClient._xsrf_cookie_name
            ],
        }
    )

  class BearerAuth(requests.auth.AuthBase):

    def __init__(self, token):
      self.token = token

    def __call__(self, r):
      r.headers['Authorization'] = f'Bearer {self.token}'
      return r

  def _try_fill_auth(self):
    if self._signed_in is not None:
      return

    api_token = os.getenv('KAGGLE_API_TOKEN')
    if api_token is not None:
      self._session.auth = KaggleHttpClient.BearerAuth(api_token)
      self._signed_in = True
      return

    if self._username and self._password:
      apikey_creds = self._username, self._password
    else:
      apikey_creds = _get_apikey_creds()
    if apikey_creds is not None:
      self._session.auth = apikey_creds
      self._signed_in = True
      return

    self._signed_in = False

  def _get_request_url(self, request):
    return f'{self._endpoint}{request.endpoint()}'

  @staticmethod
  def make_form(fields):
    body = BytesIO()
    boundary = binascii.hexlify(os.urandom(16)).decode()
    writer = codecs.lookup('utf-8')[3]

    for field in fields.items():
      field = RequestField.from_tuples(*field)
      body.write(f'--{boundary}\r\n'.encode('latin-1'))

      writer(body).write(field.render_headers())
      data = field.data

      if isinstance(data, int):
        data = str(data)

      if isinstance(data, str):
        writer(body).write(data)
      else:
        body.write(data)

      body.write(b'\r\n')

    body.write(f'--{boundary}--\r\n'.encode('latin-1'))

    content_type = f'multipart/form-data; boundary={boundary}'

    return body.getvalue(), content_type

  @staticmethod
  def requires_form(request):
    return type(request).__name__ in REQUESTS_REQUIRING_FORMS



import json
import re
from datetime import datetime, timedelta
from google.protobuf.field_mask_pb2 import FieldMask


class ObjectSerializer(object):

  def __init__(self, to_dict_value, from_dict_value):
    self.to_dict_value = to_dict_value
    self.from_dict_value = from_dict_value


class PredefinedSerializer(ObjectSerializer):

  def __init__(self):
    """Predefined objects such as int, float etc are serialized/deserialized directly."""
    ObjectSerializer.__init__(self, lambda cls, v, _: v, lambda cls, v: v)


# Adapted from https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
_pascal_to_upper_snake_case_regex = re.compile(
    '((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))'
)


def _pascal_case_to_upper_snake_case(string):
  return _pascal_to_upper_snake_case_regex.sub(r'_\1', string).upper()


def _convert(camel_input):
  words = re.findall(r'[A-Z]?[a-z]+|[A-Z]{2,}(?=[A-Z][a-z]|\d|\W|$)|\d+', camel_input)
  return '_'.join(map(str.lower, words))


class EnumSerializer(ObjectSerializer):

  def __init__(self):
    """
    Enum objects are serialized using their ".name" field and deserialized by indexing the string in the Enum type.
    Example:
    class Foo(Enum):
       TEST = 1
    foo = Foo.TEST
    foo.name # => returns "TEST"
    Foo["TEST"] # => returns Foo.TEST enum value.
    """
    ObjectSerializer.__init__(
        self,
        lambda cls, v, _: EnumSerializer._to_str(cls, v),
        lambda cls, v: EnumSerializer._from_str(cls, v),
    )

  @staticmethod
  def _to_str(cls, v):
    # "v" corresponds to an enum instance: Example foo or Foo.Test above.
    # "cls" corresponds to the enum type Foo above.
    # enum_prefix = f'{_pascal_case_to_upper_snake_case(cls.__name__)}_'
    # if v.name.startswith(enum_prefix):
    #  return v.name
    # return f'{enum_prefix}{v.name}'
    enum_prefix = f'{_pascal_case_to_upper_snake_case(cls.__name__)}_'
    if v.name.find(enum_prefix) == 0:
      return v.name[len(enum_prefix) :].lower()
    return v.name

  @staticmethod
  def _from_str(cls, v):
    # "v" corresponds to enum string: Example "TEST" above.
    # "cls" corresponds to the enum type Foo above.
    # enum_items = {item.name: item for item in cls}
    # if v in enum_items:
    #   return enum_items[v]
    #
    # # Try with enum prefix. Example: EnvironmentType.JSON -> "ENVIRONMENT_TYPE_JSON"
    # enum_prefix = _pascal_case_to_upper_snake_case(cls.__name__)
    # if v.startswith(enum_prefix):
    #   ix_start = len(enum_prefix) + 1
    #   return enum_items[v[ix_start:]]
    #
    # return enum_items[f'{enum_prefix}_{v}']
    try:
      return cls[v]
    except KeyError:
      dct = vars(cls)
      n = v.lower()
      nn = _convert(v).lower()
      enum_prefix = _pascal_case_to_upper_snake_case(cls.__name__).lower()
      for key in dct.keys():
        k = key.lower()
        if k == n:
          return dct[key]
        if k.startswith(enum_prefix) and k.endswith(n) or k.endswith(nn):
          return dct[key]
      raise


class ListSerializer(ObjectSerializer):

  def __init__(self, item_serializer: ObjectSerializer):
    """
    Lists are serialized based on the type they contain. Since objects are generated from proto files, a list always
    contains objects of the same type, which is serialized/deserialized using "item_serializer".
    """
    ObjectSerializer.__init__(
        self,
        lambda cls, l, ignore_defaults: [
            item_serializer.to_dict_value(cls, v, ignore_defaults) for v in l
        ],
        lambda cls, l: [item_serializer.from_dict_value(cls, v) for v in l],
    )


class MapSerializer(ObjectSerializer):

  def __init__(self, item_serializer: ObjectSerializer):
    """
    Maps are serialized based on type of their values. Since maps keys are always predefined types, we don't need a
    serializer for them.
    """
    ObjectSerializer.__init__(
        self,
        lambda cls, d, ignore_defaults: {
            k: item_serializer.to_dict_value(cls, v, ignore_defaults)
            for k, v in d.items()
        },
        lambda cls, d: {
            k: item_serializer.from_dict_value(cls, v) for k, v in d.items()
        },
    )


class DateTimeSerializer(ObjectSerializer):

  def __init__(self):
    """Date times are serialized/deserialized as a string in iso format"""
    ObjectSerializer.__init__(
        self,
        lambda cls, dt, _: DateTimeSerializer._to_str(dt),
        lambda _, v: DateTimeSerializer._from_str(v),
    )

  @staticmethod
  def _to_str(dt):
    return dt.isoformat(timespec='milliseconds') + 'Z'

  @staticmethod
  def _from_str(v):
    v = v.rstrip('Z')
    fields = v.rsplit('.', maxsplit=1)
    if len(fields) == 1:
      return datetime.fromisoformat(v)
    (dt, nanos) = fields
    millis = nanos[:3]
    try:
      return datetime.fromisoformat(f'{dt}.{millis}')
    except ValueError:
      return datetime.fromisoformat(dt)  # Python 3.9, 3.10


class TimeDeltaSerializer(ObjectSerializer):

  def __init__(self):
    """Time deltas are serialized/deserialized as a string in "mm:ss" format"""
    ObjectSerializer.__init__(
        self,
        lambda cls, t, _: TimeDeltaSerializer._to_dict_value(t),
        lambda cls, v: TimeDeltaSerializer._from_dict_value(v),
    )

  @staticmethod
  def _to_dict_value(delta):
    seconds = int(delta.total_seconds())
    minutes = seconds // 60
    seconds -= minutes * 60
    return '{}:{:02}'.format(int(minutes), int(seconds))

  @staticmethod
  def _from_dict_value(value):
    (minutes, seconds) = value.split(':')
    return timedelta(minutes=int(minutes), seconds=int(seconds))


class FieldMaskSerializer(ObjectSerializer):

  def __init__(self):
    """Field masks are serialized/deserialized as a string that contains a list of paths with a comma delimiter"""
    ObjectSerializer.__init__(
        self,
        lambda cls, m, _: m.ToJsonString(),
        lambda cls, v: FieldMaskSerializer._from_joined_paths(v),
    )

  @staticmethod
  def _from_joined_paths(joined_paths):
    mask = FieldMask()
    mask.FromJsonString(joined_paths)
    return mask


class KaggleObjectSerializer(ObjectSerializer):

  def __init__(self):
    """
    Kaggle objects (i.e., proto-generated types that inherit from KaggleObject) have custom "to_dict" and "from_dict"
    methods that serialize/deserialize them to/from dictionaries.
    """
    ObjectSerializer.__init__(
        self,
        # "v" is an instance of a KaggleObject. For example: "req = ListCompetitionsRequest()".
        # So "req.to_dict()" returns a dictionary with keys as json field names. Example:
        # '{"pageSize": 10, "page": 2}'
        lambda cls, v, ignore_defaults: cls.to_dict(v, ignore_defaults),
        # "cls" is the type of a KaggleObject. For example: ListCompetitionsRequest. All
        # generated Kaggle objects have "from_dict" class method that takes a dict to create a
        # new instance of the object. See "KaggleObject" class definition below.
        lambda cls, v: cls.from_dict(v),
    )


class FieldMetadata(object):

  def __init__(
      self,
      json_name,
      field_name,
      private_field_name,
      field_type,
      default_value,
      serializer,
      optional=False,
  ):
    self.json_name = json_name
    self.field_name = field_name
    self.private_field_name = private_field_name
    self.field_type = field_type
    self.default_value = default_value
    self.serializer = serializer
    self.optional = optional

  def get_as_dict_item(self, instance, ignore_defaults=True):
    value = getattr(instance, self.private_field_name)
    if ignore_defaults and value == self.default_value:
      return None
    if value is None:
      return None
    return self.serializer.to_dict_value(self.field_type, value, ignore_defaults)

  def set_from_dict(self, instance, json_dict):
    if self.json_name not in json_dict:
      return  # Ignore unknown fields
    value = json_dict[self.json_name]
    if value == self.default_value:
      return  # Ignore default values
    try:
      setattr(
          instance,
          self.private_field_name,
          self.serializer.from_dict_value(self.field_type, value),
      )
    except Exception as e:
      raise


class KaggleObject(object):

  def endpoint(self):
    raise 'Error: endpoint must be defined by the request object'

  @staticmethod
  def endpoint_path():
    return None

  @staticmethod
  def body_fields():
    return None

  @classmethod
  def prepare_from(cls, http_response):
    return cls.from_json(http_response.text)

  @staticmethod
  def method():
    return 'GET'

  def _freeze(self):
    self._is_frozen = True

  def __setattr__(self, key, value):
    if hasattr(self, '_is_frozen') and not hasattr(self, key):
      raise AttributeError(f'Unknown field for {self.__class__.__name__}: {key}')
    object.__setattr__(self, key, value)

  def to_dict(self, ignore_defaults=True):
    kv_pairs = [
        (field.json_name, field.get_as_dict_item(self, ignore_defaults))
        for field in self._fields
    ]
    return {k: v for (k, v) in kv_pairs if not ignore_defaults or v is not None}

  @staticmethod
  def to_field_map(self, ignore_defaults=True):
    kv_pairs = [
        (field.field_name, field.get_as_dict_item(self, ignore_defaults))
        for field in self._fields
    ]
    return {k: v for (k, v) in kv_pairs if not ignore_defaults or v is not None}

  @classmethod
  def from_dict(cls, json_dict):
    instance = cls()
    for field in cls._fields:
      field.set_from_dict(instance, json_dict)
    return instance

  @classmethod
  def from_json(cls, json_str):
    return cls.from_dict(json.loads(json_str))

  @staticmethod
  def to_json(self, ignore_defaults=True):
    return json.dumps(KaggleObject.to_dict(self, ignore_defaults))

  def __str__(self):
    return KaggleObject.to_json(self, ignore_defaults=False)

  def __repr__(self):
    return KaggleObject.to_json(self, ignore_defaults=False)

  def __contains__(self, field_name):
    try:
      field = self._get_field(field_name)
    except ValueError:
      return False
    value = getattr(self, field.private_field_name)
    if field.optional:
      return value is not None
    else:
      return value != field.default_value

  def __delattr__(self, field_name):
    field = self._get_field(field_name)
    setattr(self, field.private_field_name, field.default_value)

  def _get_field(self, field_name):
    field = next((f for f in self._fields if f.field_name == field_name), None)
    if field is None:
      raise ValueError(
          f'Protocol message {self.__class__.__name__} has no "{field_name}" field.'
      )
    return field



# coding=utf-8
import json
import unittest
import os
import sys
import time

from requests import HTTPError

# from kaggle.rest import ApiException
from kagglesdk.datasets.types.dataset_api_service import ApiDownloadDatasetRequest

sys.path.insert(0, '..')

sys.path.insert(0, '..')

from kaggle import api
ApiException = IOError
# Unit test names include a letter to sort them in run order.
# That seemed easier and more obvious than defining a test suite.

# Run these tests in the container with the server running.
# In the case of unresolvable failure, run reset_database.sh.

# To run from Rider, create a Python tests>Unittests run config.
# Give it the module: unit_tests.TestKaggleApi
# Set the working directory to: kaggle-api/tests
# Define some envars:
# KAGGLE_API_ENDPOINT=http://localhost
# KAGGLE_API_ENVIRONMENT=LOCALHOST
# KAGGLE_CONFIG_DIR=/home/kaggle/.config/kaggle/dev
# KAGGLE_KEY=local_api_token
# KAGGLE_USERNAME=<kaggle-user-name>

# Add those envars to the Python Tests>Unittest template to
# make running individual tests easier.

test_user = api.config_values['username']
model_title = 'testing'
instance_name = 'test'
framework_name = 'jax'
kernel_name = 'testing'
dataset_name = 'kaggleapi-testdataset'
up_file = 'sample_submission.csv'
description = 'House prices submission message'
competition = 'house-prices-advanced-regression-techniques'
dataset_directory = 'dataset'
kernel_directory = 'kernel'
model_directory = 'model'
model_inst_directory = os.path.join(model_directory, 'instance')
model_inst_vers_directory = os.path.join(model_inst_directory, 'version')

# Max retries to get kernel status
max_status_tries = 10


def tearDownModule():
  file = os.path.join(dataset_directory, api.DATASET_METADATA_FILE)
  if os.path.exists(file):
    os.remove(file)
  file = os.path.join(kernel_directory, api.KERNEL_METADATA_FILE)
  if os.path.exists(file):
    os.remove(file)
  file = os.path.join(model_directory, api.MODEL_METADATA_FILE)
  if os.path.exists(file):
    os.remove(file)
  file = os.path.join(model_inst_directory, api.MODEL_INSTANCE_METADATA_FILE)
  if os.path.exists(file):
    os.remove(file)


def update_kernel_metadata_file(metadata_file, k_name):
  with open(metadata_file) as f:
    meta_data = json.load(f)
    meta_id = meta_data['id']
    if 'INSERT_KERNEL_SLUG_HERE' in meta_id:
      meta_id = meta_id.replace('INSERT_KERNEL_SLUG_HERE', k_name)
    meta_title = meta_data['title']
    if 'INSERT_TITLE_HERE' == meta_title:
      meta_title = k_name
    meta_path = meta_data['code_file']
    if 'INSERT_CODE_FILE_PATH_HERE' == meta_path:
      meta_path = f'{k_name}.ipynb'
    meta_data['id'] = meta_id
    meta_data['title'] = meta_title
    meta_data['code_file'] = meta_path
    meta_data['language'] = 'python'
    meta_data['kernel_type'] = 'notebook'
  with open(metadata_file, 'w') as f:
    json.dump(meta_data, f, indent=2)
  return meta_data


def initialize_dataset_metadata_file(dataset_dir):
  metadata_file = os.path.join(dataset_dir, api.DATASET_METADATA_FILE)
  try:
    with open(metadata_file) as f:
      original = json.load(f)
      if 'versionNumber' in original:
        version_num = int(original['versionNumber'])
      else:
        version_num = 0
  except FileNotFoundError:
    version_num = 0
  version_num += 1
  return version_num, metadata_file


def update_dataset_metadata_file(metadata_file, data_name, version_num):
  with open(metadata_file) as f:
    meta_data = json.load(f)
    meta_id = meta_data['id']
    if 'INSERT_SLUG_HERE' in meta_id:
      meta_id = meta_id.replace('INSERT_SLUG_HERE', data_name)
    meta_title = meta_data['title']
    if 'INSERT_TITLE_HERE' == meta_title:
      meta_title = data_name
    meta_data['id'] = meta_id
    meta_data['title'] = meta_title
    meta_data['versionNumber'] = version_num
  if not 'resources' in meta_data:
    resource_list = [{
        "path": "data.csv",
        "description": "Description",
        "schema": {
            "fields": [{
                "name": "NumberField",
                "description": "id",
                "type": "number"
            }, {
                "name": "StringField",
                "description": "label",
                "type": "string"
            }]
        }
    }]
    meta_data.update({'resources': resource_list})
  with open(metadata_file, 'w') as f:
    json.dump(meta_data, f)


def update_model_metadata(metadata_file, owner, title, slug):
  with open(metadata_file) as f:
    meta_data = json.load(f)
    meta_id = meta_data['ownerSlug']
    if 'INSERT_OWNER_SLUG_HERE' == meta_id:
      meta_id = owner
    meta_title = meta_data['title']
    if 'INSERT_TITLE_HERE' == meta_title:
      meta_title = title
    meta_path = meta_data['slug']
    if 'INSERT_SLUG_HERE' == meta_path:
      meta_path = slug
    meta_data['ownerSlug'] = meta_id
    meta_data['title'] = meta_title
    meta_data['slug'] = meta_path
  with open(metadata_file, 'w') as f:
    json.dump(meta_data, f, indent=2)
  return meta_data


def update_model_instance_metadata(metadata_file, owner, model_slug,
                                   instance_slug, framework):
  with open(metadata_file) as f:
    meta_data = json.load(f)
    meta_owner = meta_data['ownerSlug']
    if 'INSERT_OWNER_SLUG_HERE' == meta_owner:
      meta_owner = owner
    meta_framework = meta_data['framework']
    if 'INSERT_FRAMEWORK_HERE' == meta_framework:
      meta_framework = framework
    meta_instance = meta_data['instanceSlug']
    if 'INSERT_INSTANCE_SLUG_HERE' == meta_instance:
      meta_instance = instance_slug
    meta_model = meta_data['modelSlug']
    if 'INSERT_EXISTING_MODEL_SLUG_HERE' == meta_model:
      meta_model = model_slug
    meta_data['ownerSlug'] = meta_owner
    meta_data['modelSlug'] = meta_model
    meta_data['framework'] = meta_framework
    meta_data['instanceSlug'] = meta_instance
  with open(metadata_file, 'w') as f:
    json.dump(meta_data, f, indent=2)
  return meta_data


def print_fields(instance, fields):  # For debugging.
  for f in fields:
    if not hasattr(instance, api.camel_to_snake(f)):
      print(f"Missing field: {f} named: {api.camel_to_snake(f)}")


class TestKaggleApi(unittest.TestCase):

  version_number, meta_file = initialize_dataset_metadata_file(
      dataset_directory)

  # Initialized from Response objects.
  competition_file = None
  kernel_slug = ''
  kernel_metadata_path = ''
  dataset = ''
  dataset_file = None
  model_instance = ''
  model_meta_data = None
  model_metadata_file = ''
  instance_metadata_file = ''

  # Inbox

  def test_files_upload(self):
    a = 2 # Change this value to run this test.
    if a - 1 == 1:
      return # Only run this test when needed because it uploads an inbox file.
    filename = 'tmp_file.test'
    with open(filename, 'w') as f:
      f.write('test')
    try:
      api.files_upload_cli([filename], 'kaggle-api-test', False, False)
    finally:
      if os.path.exists('tmp_file.test'):
        os.remove('tmp_file.test')

  # Kernels

  def test_kernels_a_list(self):
    try:
      kernels = api.kernels_list(sort_by='dateCreated', user='stevemessick', language='python')
      self.assertGreater(len(kernels), 0)
      api.kernels_list_cli(user='stevemessick', csv_display=True)
    except ApiException as e:
      self.fail(f"kernels_list failed: {e}")

  def test_kernels_b_initialize(self):
    try:
      self.kernel_metadata_path = api.kernels_initialize(kernel_directory)
      self.assertTrue(os.path.exists(self.kernel_metadata_path))
    except ApiException as e:
      self.fail(f"kernels_initialize failed: {e}")

  def test_kernels_c_push(self):
    if self.kernel_metadata_path == '':
      self.test_kernels_b_initialize()
    try:
      md = update_kernel_metadata_file(self.kernel_metadata_path, kernel_name)
      push_result = api.kernels_push(kernel_directory)
      self.assertIsNotNone(push_result.ref)
      self.assertTrue(isinstance(push_result.version_number, int))
      self.kernel_slug = md['id']
      time.sleep(30)
    except ApiException as e:
      self.fail(f"kernels_push failed: {e}")

  def test_kernels_d_status(self):
    if self.kernel_slug == '':
      self.test_kernels_c_push()
    try:
      status_result = api.kernels_status(self.kernel_slug)
      start_time = time.time()
      # If this loop is stuck because the kernel stays queued, go to the Kaggle website
      # on localhost and cancel the active event. That will exit the loop, but you may
      # need to clean up other active kernels to get it to run again.
      count = 0
      while status_result.status == 'running' or status_result.status == 'queued' or count >= max_status_tries:
        time.sleep(5)
        status_result = api.kernels_status(self.kernel_slug)
        print(status_result.status)
      if count >= max_status_tries:
        self.fail(f"Could not get kernel status in allowed trys. Status: {status_result.status}")
      end_time = time.time()
      print(f'kernels_status ready in {end_time-start_time}s')
    except ApiException as e:
      self.fail(f"kernels_status failed: {e}")

  def test_kernels_e_list_files(self):
    if self.kernel_slug == '':
      self.test_kernels_c_push()
    try:
      fs = api.kernels_list_files(self.kernel_slug)
      # TODO Make sure the test uses a kernel that has at least one file.
      self.assertGreaterEqual(len(fs.files), 0)
    except ApiException as e:
      self.fail(f"kernels_list_files failed: {e}")

  def test_kernels_f_output(self):
    fs = []
    try:
      fs, token = api.kernels_output('kerneler/sqlite-global-default', 'kernel/tmp')
      self.assertIsInstance(fs, list)
      if token:
        print(token)
    except ApiException as e:
      self.fail(f"kernels_output failed: {e}")
    finally:
      for file in fs:
        if os.path.exists(file):
          os.remove(file)
      if os.path.exists('kernel/tmp'):
        os.rmdir('kernel/tmp')

  def test_kernels_g_pull(self):
    if self.kernel_metadata_path == '':
      self.test_kernels_c_push()
    fs = ''
    try:
      fs = api.kernels_pull(f'{test_user}/testing', 'kernel/tmp', metadata=True)
      self.assertTrue(os.path.exists(fs))
      with open(f'{fs}/{self.kernel_metadata_path.split("/")[1]}') as f:
        metadata = json.load(f)
        [
          self.assertTrue(metadata.get(f))
          for f in ['id','id_no','title','code_file','language','kernel_type']
        ]
        [
          self.assertTrue(metadata.get(f) is not None)
          for f in ['is_private','enable_gpu','enable_tpu','enable_internet','keywords','dataset_sources','kernel_sources','competition_sources','model_sources']
        ]
    except ApiException as e:
      self.fail(f"kernels_pull failed: {e}")
    finally:
      for file in [
          f'{fs}/{self.kernel_metadata_path.split("/")[1]}',
          f'{fs}/{kernel_name}.ipynb'
      ]:
        if os.path.exists(file):
          os.remove(file)
      if os.path.exists(fs):
        os.rmdir(fs)

  # Competitions

  def test_competition_a_list(self):
    try:
      competitions = api.competitions_list(group='general')
      self.assertGreater(len(competitions), 0)
      self.assertLessEqual(len(competitions), 20)
      [
          self.assertTrue(hasattr(competitions[0], api.camel_to_snake(f)))
          for f in api.competition_fields
      ]
      competitions = api.competitions_list(page=2, category='gettingStarted', sort_by='prize')
      self.assertEqual(len(competitions), 0)
    except ApiException as e:
      self.fail(f"competitions_list failed: {e}")

  def test_competition_b_submit(self):
    try:
      self.skip_submissions = False
      result = api.competition_submit_cli(up_file, description, competition)
      if not result or not result.startswith('Successfully submitted'):
        self.fail(f'competition_submit failed: {result}')
    except HTTPError:
      # Handle submission limit reached gracefully (potentially skip the test)
      print('Competition submission limit reached for the day')
      self.skip_submissions = True
      pass
    except ApiException as e:
      self.fail(f"competition_submit failed: {e}")

  def test_competition_c_submissions(self):
    self.test_competition_b_submit()
    try:
      submissions = api.competition_submissions(competition)
      self.assertIsInstance(submissions, list)
      if not self.skip_submissions:
        self.assertGreater(len(submissions), 0)
        [
            self.assertTrue(hasattr(submissions[0], api.camel_to_snake(f)))
            for f in api.submission_fields
        ]
    except ApiException as e:
      self.fail(f"competition_submissions failed: {e}")

  def test_competition_d_list_files(self):
    try:
      competition_files = api.competition_list_files(competition).files
      self.assertIsInstance(competition_files, list)
      self.assertGreater(len(competition_files), 0)
      self.competition_file = competition_files[0]
      [
          self.assertTrue(hasattr(competition_files[0], api.camel_to_snake(f)))
          for f in api.competition_file_fields
      ]
    except ApiException as e:
      self.fail(f"competition_list_files failed: {e}")

  def test_competition_e_download_file(self):
    if self.competition_file is None:
      self.test_competition_d_list_files()
    try:
      api.competition_download_file(
          competition, self.competition_file.ref, force=True)
      self.assertTrue(os.path.exists(self.competition_file.ref))
      api.competition_download_file(
          competition, self.competition_file.ref, force=False)
    except ApiException as e:
      self.fail(f"competition_download_file failed: {e}")
    finally:
      if os.path.exists(self.competition_file.ref):
        os.remove(self.competition_file.ref)

  def test_competition_f_download_files(self):
    try:
      api.competition_download_files(competition)
      self.assertTrue(os.path.exists(f'{competition}.zip'))
      self.assertTrue(os.path.getsize(f'{competition}.zip') > 0)
    except ApiException as e:
      self.fail(f"competition_download_files failed: {e}")
    finally:
      if os.path.exists(f'{competition}.zip'):
        os.remove(f'{competition}.zip')

  def test_competition_g_leaderboard_view(self):
    try:
      result = api.competition_leaderboard_view(competition)
      self.assertIsInstance(result, list)
      self.assertGreater(len(result), 0)
      [
          self.assertTrue(hasattr(result[0], api.camel_to_snake(f)))
          for f in api.competition_leaderboard_fields
      ]
    except ApiException as e:
      self.fail(f"competition_leaderboard_view failed: {e}")

  def test_competition_h_leaderboard_download(self):
    try:
      api.competition_leaderboard_download(competition, 'tmp')
      self.assertTrue(os.path.exists(f'tmp/{competition}.zip'))
    except ApiException as e:
      self.fail(f"competition_leaderboard_download failed: {e}")
    finally:
      if os.path.exists(f'tmp/{competition}.zip'):
        os.remove(f'tmp/{competition}.zip')
      if os.path.exists('tmp'):
        os.rmdir('tmp')

  # Datasets

  def test_dataset_a_list(self):
    try:
      datasets = api.dataset_list(sort_by='votes')
      self.assertGreater(len(datasets), 0)
      self.dataset = str(datasets[0].ref)
      [
          self.assertTrue(hasattr(datasets[0], api.camel_to_snake(f)))
          for f in api.dataset_fields
      ]
      datasets = api.dataset_list(license_name='other', file_type='bigQuery')
      self.assertGreater(len(datasets), 10)
    except ApiException as e:
      self.fail(f"dataset_list failed: {e}")

  def test_dataset_b_metadata(self):
    if self.dataset == '':
      self.test_dataset_a_list()
    m = ''
    try:
      m = api.dataset_metadata(self.dataset, dataset_directory)
      self.assertTrue(os.path.exists(m))
    except ApiException as e:
      self.fail(f"dataset_metadata failed: {e}")

  def test_dataset_c_metadata_update(self):
    if self.dataset == '':
      self.test_dataset_a_list()
    if not os.path.exists(
        os.path.join(dataset_directory, api.DATASET_METADATA_FILE)):
      self.test_dataset_b_metadata()
    try:
      api.dataset_metadata_update(self.dataset, dataset_directory)
      # TODO Make the API method return something, and not exit when it fails.
    except ApiException as e:
      self.fail(f"dataset_metadata_update failed: {e}")

  def test_dataset_d_list_files(self):
    if self.dataset == '':
      self.test_dataset_a_list()
    try:
      response = api.dataset_list_files(self.dataset)
      self.assertIsInstance(response.dataset_files, list)
      self.assertGreater(len(response.dataset_files), 0)
      self.dataset_file = response.dataset_files[0]
      [
          self.assertTrue(hasattr(self.dataset_file, api.camel_to_snake(f)))
          for f in api.dataset_file_fields
      ]
      api.dataset_list_files_cli(self.dataset)
    except ApiException as e:
      self.fail(f"dataset_list_files failed: {e}")

  def test_dataset_e_status(self):
    if self.dataset == '':
      self.test_dataset_a_list()
    try:
      status = api.dataset_status(self.dataset)
      self.assertIn(status, ['ready', 'pending', 'error'])
    except ApiException as e:
      self.fail(f"dataset_status failed: {e}")

  def test_dataset_f_download_file(self):
    if self.dataset_file is None:
      self.test_dataset_d_list_files()
    try:
      api.dataset_download_file(self.dataset, self.dataset_file.name, 'tmp')
      self.assertTrue(os.path.exists(f'tmp/{self.dataset_file.name}'))
    except ApiException as e:
      self.fail(f"dataset_download_file failed: {e}")
    finally:
      if os.path.exists(f'tmp/{self.dataset_file.name}'):
        os.remove(f'tmp/{self.dataset_file.name}')
      if os.path.exists('tmp'):
        os.rmdir('tmp')

  def test_dataset_g_download_files(self):
    if self.dataset == '':
      self.test_dataset_a_list()
    ds = ['a', 'b']
    try:
      api.dataset_download_files(self.dataset)
      ds = self.dataset.split('/')
      self.assertTrue(os.path.exists(f'{ds[1]}.zip'))
    except ApiException as e:
      self.fail(f"dataset_download_files failed: {e}")
    finally:
      if os.path.exists(f'{ds[1]}.zip'):
        os.remove(f'{ds[1]}.zip')

  def test_dataset_h_initialize(self):
    try:
      api.dataset_initialize('dataset')
      self.assertTrue(
          os.path.exists(
              os.path.join(dataset_directory, api.DATASET_METADATA_FILE)))
    except ApiException as e:
      self.fail(f"dataset_initialize failed: {e}")

  def test_dataset_i_create_new(self):
    if not os.path.exists(
        os.path.join(dataset_directory, api.DATASET_METADATA_FILE)):
      self.test_dataset_h_initialize()
    try:
      update_dataset_metadata_file(self.meta_file, dataset_name,
                                   self.version_number)
      new_dataset = api.dataset_create_new(dataset_directory)
      self.assertIsNotNone(new_dataset)
      if new_dataset.error is not None:
        if 'already in use' in new_dataset.error:
          print(new_dataset.error)  # This is likely to happen, and that's OK.
          self.skip_create_version = True
        else:
          self.fail(f"dataset_create_new failed: {new_dataset.error}")
    except ApiException as e:
      self.fail(f"dataset_create_new failed: {e}")

  def test_dataset_j_create_version(self):
    if not os.path.exists(
        os.path.join(dataset_directory, api.DATASET_METADATA_FILE)):
      self.test_dataset_i_create_new()
    try:
      new_version = api.dataset_create_version(dataset_directory, "Notes")
      self.assertIsNotNone(new_version)
      self.assertTrue(new_version.error == '')
      self.assertFalse(new_version.ref == '')
    except ApiException as e:
      self.fail(f"dataset_create_version failed: {e}")

  # Models

  def test_model_a_list(self):
    try:
      ms = api.model_list()
      self.assertIsInstance(ms, list)
      self.assertGreater(len(ms), 0)
      [self.assertTrue(hasattr(ms[0], api.camel_to_snake(f)))
        for f in api.model_fields]
    except ApiException as e:
      self.fail(f"models_list failed: {e}")

  def test_model_b_initialize(self):
    try:
      self.model_metadata_file = api.model_initialize(model_directory)
      self.assertTrue(os.path.exists(self.model_metadata_file))
      self.model_meta_data = update_model_metadata(self.model_metadata_file,
                                                   test_user, model_title,
                                                   model_title)
      self.model_instance = f'{test_user}/{self.model_meta_data["slug"]}/{framework_name}/{instance_name}'
    except ApiException as e:
      self.fail(f"model_initialize failed: {e}")

  def test_model_c_create_new(self):
    if self.model_metadata_file == '':
      self.test_model_b_initialize()
    try:
      model = api.model_create_new(model_directory)
      if model.error:
        if 'already used' in model.error:
          delete_response = api.model_delete(f'{test_user}/{model_title}', True)
          if delete_response.error:
            self.fail(delete_response.error)
          else:
            model = api.model_create_new(model_directory)
            if model.error:
              self.fail(model.error)
        else:
          self.fail(model.error)
      self.assertIsNotNone(model.ref)
      self.assertGreater(len(model.ref), 0)
      [self.assertTrue(hasattr(model, api.camel_to_snake(f)))
       for f in ['id', 'url']]
    except ApiException as e:
      self.fail(f"model_create_new failed: {e}")

  def test_model_d_get(self):
    try:
      model_data = api.model_get(f'{test_user}/{model_title}')
      self.assertIsNotNone(model_data.ref)
      self.assertGreater(len(model_data.ref), 0)
      self.assertEqual(model_data.title, model_title)
      [self.assertTrue(hasattr(model_data, api.camel_to_snake(f)))
       for f in api.model_all_fields]
    except ApiException as e:
      self.fail(f"model_get failed: {e}")

  def test_model_e_update(self):
    if self.model_metadata_file == '':
      self.test_model_c_create_new()
    try:
      update_response = api.model_update(model_directory)
      self.assertEqual(len(update_response.error), 0)   
      self.assertIsNotNone(update_response.ref)
      self.assertGreater(len(update_response.ref), 0)
    except ApiException as e:
      self.fail(f"model_update failed: {e}")

  # Model instances

  def test_model_instance_a_initialize(self):
    try:
      self.instance_metadata_file = api.model_instance_initialize(
          model_inst_directory)
      self.assertTrue(os.path.exists(self.instance_metadata_file))
    except ApiException as e:
      self.fail(f"model_instance_initialize failed: {e}")

  def test_model_instance_b_create(self, check_result: bool = True):
    if self.model_meta_data is None:
      self.test_model_b_initialize()
    if self.instance_metadata_file == '':
      self.test_model_instance_a_initialize()
    try:
      update_model_instance_metadata(self.instance_metadata_file, test_user,
                                     self.model_meta_data['slug'],
                                     instance_name, framework_name)
      inst_create_resp = api.model_instance_create(model_inst_directory)
      if check_result:
        if inst_create_resp.error:
          if 'already exists' in inst_create_resp.error:
            delete_response = api.model_instance_delete(f'{test_user}/{model_title}', True)
            if delete_response.error:
              self.fail(delete_response.error)
            else:
              inst_create_resp = api.model_instance_create(model_inst_directory)
              if inst_create_resp.error:
                self.fail(inst_create_resp.error)
          else:
            self.fail(inst_create_resp.error)
        self.assertIsNotNone(inst_create_resp.ref)
        self.assertGreater(len(inst_create_resp.ref), 0)
    except ApiException as e:
      self.fail(f"model_instance_create failed: {e}")

  def test_model_instance_b_wait_after_create(self):
    # When running all tests sequentially, give the new model some time to stabilize.
    time.sleep(10)  # TODO: Find a better way to detect model stability.

  def test_model_instance_c_get(self):
    if self.model_instance == '':
      self.test_model_b_initialize()
    try:
      inst_get_resp = api.model_instance_get(self.model_instance)
      self.assertIsNotNone(inst_get_resp.url)
      self.assertGreater(len(inst_get_resp.url), 0)
      os.makedirs('model/tmp', exist_ok=True)
      api.model_instance_get_cli(self.model_instance, 'model/tmp')
    except ApiException as e:
      self.fail(f"model_instance_get failed: {e}")
    finally:
      os.remove('model/tmp/model-instance-metadata.json')
      os.rmdir('model/tmp')

  def test_model_instance_d_files(self):
    if self.model_instance == '':
      self.test_model_b_initialize()
    try:
      inst_files_resp = api.model_instance_files(self.model_instance)
      self.assertIsInstance(inst_files_resp.files, list)
      self.assertGreater(len(inst_files_resp.files), 0)
      [self.assertTrue(hasattr(inst_files_resp.files[0], api.camel_to_snake(f)))
       for f in api.model_file_fields]
    except ApiException as e:
      self.fail(f"model_instance_files failed: {e}")

  def test_model_instance_e_update(self):
    if self.model_instance == '':
      self.test_model_b_initialize()
      self.test_model_instance_a_initialize()
      try:
        self.test_model_c_create_new()
      except AssertionError:
        pass
      try:
        update_model_instance_metadata(self.instance_metadata_file, test_user,
                                     self.model_meta_data['slug'],
                                     instance_name, framework_name)
      except AssertionError:
        pass
      self.test_model_instance_b_create(check_result=False)
      self.test_model_instance_b_wait_after_create()
    try:
      inst_update_resp = api.model_instance_update(model_inst_directory)
      self.assertIsNotNone(inst_update_resp)
      self.assertIsNotNone(inst_update_resp.ref)
      self.assertGreater(len(inst_update_resp.ref), 0)
      [self.assertTrue(hasattr(inst_update_resp, api.camel_to_snake(f)))
       for f in ['error', 'id', 'ref', 'url']]
    except ApiException as e:
      self.fail(f"model_instance_update failed: {e}")

  # Model instance versions

  def test_model_instance_version_a_create(self):
    if self.model_instance == '':
      self.test_model_b_initialize()
    try:
      version_metadata_resp = api.model_instance_version_create(
          self.model_instance, model_inst_vers_directory)
      self.assertIsNotNone(version_metadata_resp.ref)
      [self.assertTrue(hasattr(version_metadata_resp, api.camel_to_snake(f)))
       for f in ['id', 'url', 'error']]
    except ApiException as e:
      self.fail(f"model_instance_version_create failed: {e}")

  def test_model_instance_version_b_files(self):
    if self.model_instance == '':
      self.test_model_b_initialize()
    try:
      r = api.model_instance_version_files(f'{self.model_instance}/1')
      self.assertIsInstance(r.files, list)
      self.assertGreater(len(r.files), 0)
      api.model_instance_version_files_cli(f'{self.model_instance}/1')
    except ApiException as e:
      self.fail(f"model_instance_version_files failed: {e}")

  def test_model_instance_version_c_download(self):
    if self.model_instance == '':
      self.test_model_b_initialize()
    version_file = ''
    try:
      version_file = api.model_instance_version_download(
          f'{self.model_instance}/1', 'tmp', force=True)
      self.assertTrue(os.path.exists(version_file))
    except KeyError:
      pass  # TODO Create a version that has content.
    except ApiException as e:
      self.fail(f"model_instance_version_download failed: {e}")
    finally:
      if os.path.exists(version_file):
        os.remove(version_file)
      if os.path.exists('tmp'):
        os.rmdir('tmp')

  # Model deletion

  def test_model_instance_version_d_delete(self):
    if self.model_instance == '':
      self.test_model_b_initialize()
    try:
      version_delete_resp = api.model_instance_version_delete(
          f'{self.model_instance}/1', True)
      self.assertEqual(len(version_delete_resp.error), 0, msg=version_delete_resp.error)
    except ApiException as e:
      self.fail(f"model_instance_version_delete failed: {e}")

  def test_model_instance_x_delete(self):
    if self.model_instance == '':
      self.test_model_b_initialize()
    try:
      inst_update_resp = api.model_instance_delete(self.model_instance, True)
      self.assertIsNotNone(inst_update_resp)
      if len(inst_update_resp.error):
        print(inst_update_resp.error)
      self.assertEqual(len(inst_update_resp.error), 0)
    except ApiException as e:
      self.fail(f"model_instance_delete failed: {e}")

  def test_model_z_delete(self):
    try:
      delete_response = api.model_delete(f'{test_user}/{model_title}', True)
      if delete_response.error:
        self.fail(delete_response.error)
      else:
        pass
    except ApiException as e:
      self.fail(f"model_delete failed: {e}")


if __name__ == '__main__':
  unittest.main()
